"""\
Copyright (c) 2023, Flagstaff Solutions, LLC
All rights reserved.

"""
import json
import logging
import os
import re
import subprocess
import sys
from abc import ABC
from urllib.parse import unquote, urlparse

import git
from IPython import get_ipython

from gofigr.models import CodeLanguage
from gofigr.databricks import get_dbutils

PATH_WARNING = "To fix this warning, you can manually specify the notebook name & path in the call to configure(). " \
               "Please see https://gofigr.io/docs/gofigr-python/latest/customization.html#notebook-name-path " \
               "for details."


class Annotator(ABC):
    """\
    Annotates figure revisions with pertinent information, such as cell code, variable values, etc.

    """
    def annotate(self, revision):
        """\
        Annotates a figure revision in-place.

        :param revision: revision to annotate
        :return: annotated revision (same object as input).

        """
        raise NotImplementedError("Must be implemented in subclass")


class IPythonAnnotator(Annotator):
    """\
    Annotates figures within the IPython/Jupyter environment.

    """
    def get_ip_extension(self):
        """\
        :return: IPython extension if available, None otherwise.

        """
        try:
            get_extension = get_ipython().user_ns.get("get_extension")
            return get_extension() if get_extension is not None else None
        except Exception as e:  # pylint: disable=broad-exception-caught
            logging.debug(f"IPython extension could not be found: {e}")
            return None

    def annotate(self, revision):
        """
        Annotates the figure revision.

        :param revision: FigureRevision
        :return: annotated FigureRevision

        """
        ext = self.get_ip_extension()
        if ext is None:
            return revision
        else:
            return self.annotate_ip(revision, ext)

    def annotate_ip(self, revision, ext):
        """\
        Annotates the figure revision assuming the IPython extension is available.

        :param revision: GoFigr revision
        :param ext: Jupyter extension
        :return: annotated revision
        """
        raise NotImplementedError("Must be implemented in subclass")


class CellIdAnnotator(IPythonAnnotator):
    """Annotates revisions with the ID of the Jupyter cell"""
    def annotate_ip(self, revision, ext):
        if revision.metadata is None:
            revision.metadata = {}

        try:
            cell_id = ext.cell.cell_id
        except AttributeError as e:
            logging.debug(e)
            cell_id = None

        revision.metadata['cell_id'] = cell_id

        return revision


class CellCodeAnnotator(IPythonAnnotator):
    """Annotates revisions with cell contents"""
    def annotate_ip(self, revision, ext):
        if ext.cell is not None:
            code = ext.cell.raw_cell
        else:
            code = "N/A"

        revision.data.append(revision.client.CodeData(name="Jupyter Cell",
                                                      language=CodeLanguage.PYTHON,
                                                      contents=code))
        return revision


class PipFreezeAnnotator(Annotator):
    """Annotates revisions with the output of pip freeze"""
    def __init__(self, cache=True):
        """\
        :param cache: if True, will only run pip freeze once and cache the output
        """
        super().__init__()
        self.cache = cache
        self.cached_output = None

    def annotate(self, revision):
        if self.cache and self.cached_output:
            output = self.cached_output
        else:
            try:
                output = subprocess.check_output(["pip", "freeze"]).decode('ascii')
                self.cached_output = output
            except subprocess.CalledProcessError as e:
                output = e.output

        revision.data.append(revision.client.TextData(name="pip freeze", contents=output))
        return revision


class ScriptAnnotator(Annotator):
    """Annotates revisions with the code of the running script"""
    def annotate(self, revision):
        if get_ipython() is not None:  # skip if running interactively
            return revision

        if os.path.exists(sys.argv[0]):
            revision.data.append(revision.client.FileData.read(sys.argv[0]))

        if revision.metadata is None:
            revision.metadata = {}

        revision.metadata['argv'] = list(sys.argv)
        return revision


class GitAnnotator(Annotator):
    """Annotates revisions with Git information"""

    def annotate(self, revision):
        """
        Generates a link for the current commit of a local repository.
        """
        try:
            # 1. Initialize the repository object
            repo = git.Repo(".", search_parent_directories=True)

            try:
                branch_name = repo.active_branch.name
            except TypeError:
                branch_name = "Detached HEAD"

            # 2. Get the current commit hash
            commit_hash = repo.head.commit.hexsha

            # 3. Get the URL of the 'origin' remote
            remote_url = repo.remotes.origin.url

            # 4. Clean and reformat the remote URL to a standard HTTPS format
            #    Handles both SSH (git@github.com:user/repo.git) and HTTPS formats
            http_url = re.sub(r'\.git$', '', remote_url)  # Remove .git suffix
            if http_url.startswith('git@'):
                # Convert SSH URL to HTTPS
                http_url = http_url.replace(':', '/').replace('git@', 'https://', 1)

            # 5. Construct the final commit link
            if 'github.com' in http_url.lower():
                commit_link = f"{http_url}/commit/{commit_hash}"
            elif 'bitbucket.org' in http_url.lower():
                commit_link = f"{http_url}/commits/{commit_hash}"
            else:
                commit_link = None

            revision.metadata['git'] = {'branch': branch_name,
                                        'hash': commit_hash,
                                        'remote_url': remote_url,
                                        'commit_link': commit_link}

        except git.exc.InvalidGitRepositoryError:
            logging.debug("Error: Not a valid git repository.")
        except Exception as e:  # pylint: disable=broad-exception-caught
            logging.debug(f"An unexpected error occurred: {e}")
            return None

        return revision


class SystemAnnotator(Annotator):
    """Annotates revisions with the OS version"""
    def annotate(self, revision):
        try:
            output = subprocess.check_output(["uname", "-a"]).decode('ascii')
        except subprocess.CalledProcessError as e:
            output = e.output

        revision.data.append(revision.client.TextData(name="System Info", contents=output))
        return revision


NOTEBOOK_PATH = "notebook_path"
NOTEBOOK_NAME = "notebook_name"
NOTEBOOK_URL = "url"
NOTEBOOK_KERNEL = "kernel"
PYTHON_VERSION = "python_version"
BACKEND_NAME = "backend"


_ACTIVE_TAB_TITLE = "active_tab_title"


def _parse_path_from_tab_title(title):
    """Parses out the notebook path from the tab/widget title"""
    for line in title.splitlines(keepends=False):
        m = re.match(r'Path:\s*(.*)\s*', line)
        if m:
            return m.group(1)
    return None


class NotebookMetadataAnnotator(IPythonAnnotator):
    """Annotates revisions with notebook metadata, including filename & path, as well as the full URL"""
    def parse_from_databricks(self):
        """Returns notebook path if running in Databricks"""
        try:
            # pylint: disable=undefined-variable
            context = get_dbutils(get_ipython()).notebook.entry_point.getDbutils().notebook().getContext()
            nb = context.notebookPath().get()
            return {NOTEBOOK_PATH: nb, NOTEBOOK_NAME: os.path.basename(nb)}
        except Exception as e:  # pylint: disable=broad-exception-caught
            logging.debug(f"Unable to parse notebook metadata from Databricks: {e}")
            return None

    def parse_from_vscode(self):
        """Returns notebook path if running in VSCode"""
        ext = self.get_ip_extension()
        if ext is None:
            return None

        try:
            if ext.cell is None or getattr(ext.cell, "cell_id") is None:
                return None
            elif "vscode-notebook-cell:" not in ext.cell.cell_id:
                return None

            m = re.match(r'^vscode-notebook-cell:(.*)#.*$', unquote(ext.cell.cell_id))
            if m is None:
                return None

            notebook_path = m.group(1)
            notebook_name = os.path.basename(notebook_path)

            return {NOTEBOOK_PATH: notebook_path,
                    NOTEBOOK_NAME: notebook_name}
        except Exception:  # pylint: disable=broad-exception-caught
            return None

    def try_get_metadata(self):
        """Infers the notebook path & name using currently available metadata if possible, returning None otherwise"""
        try:
            return self.parse_metadata(error=False)
        except Exception:  # pylint: disable=broad-exception-caught
            return None

    def _parse_from_comms(self, comm_data):
        if comm_data is None:
            return None

        return {NOTEBOOK_PATH: comm_data.get('notebook_path'),
                NOTEBOOK_NAME: comm_data.get('title'),
                NOTEBOOK_URL: comm_data.get('url')}

    def _parse_from_proxy(self, meta, error):
        if 'url' not in meta and _ACTIVE_TAB_TITLE not in meta:
            if error:
                raise RuntimeError("No URL found in Notebook metadata")
            else:
                return None

        notebook_name = None

        # Try parsing the name from the title first
        if _ACTIVE_TAB_TITLE in meta and meta[_ACTIVE_TAB_TITLE] is not None:
            notebook_name = _parse_path_from_tab_title(meta[_ACTIVE_TAB_TITLE])

        # If that doesn't work, try the URL
        if notebook_name is None:
            notebook_name = unquote(urlparse(meta['url']).path.rsplit('/', 1)[-1])

        notebook_dir = get_ipython().starting_dir
        full_path = None

        for candidate_path in [os.path.join(notebook_dir, notebook_name),
                               os.path.join(notebook_dir, os.path.basename(notebook_name)),
                               os.path.join(os.path.dirname(notebook_dir), notebook_name),
                               os.path.join(os.path.dirname(notebook_dir), os.path.basename(notebook_name))]:
            if os.path.exists(candidate_path):
                full_path = candidate_path
                break

        if full_path is None:
            full_path = os.path.join(notebook_dir, notebook_name)  # might still be helpful, even if slightly incorrect
            print(f"The inferred path for the notebook does not exist: {full_path}. {PATH_WARNING}", file=sys.stderr)

        return {NOTEBOOK_PATH: full_path,
                NOTEBOOK_NAME: notebook_name,
                NOTEBOOK_URL: meta.get('url')}

    def parse_metadata(self, error=True):
        """
        Infers the notebook path & name from metadata passed through the WebSocket (if available)

        :param revision: GoFigr Revision instance
        :param error: if True, will raise an error if metadata is not available
        """
        vsc_meta = self.parse_from_vscode()
        if vsc_meta is not None:
            return vsc_meta

        db_meta = self.parse_from_databricks()
        if db_meta is not None:
            return db_meta

        # At this point the metadata needs to come from the JavaScript proxy
        ext = self.get_ip_extension()
        if ext is None:
            if error:
                raise RuntimeError("No Notebook metadata available")
            else:
                return None

        if hasattr(ext, "comm_data"):
            return self._parse_from_comms(ext.comm_data)
        else:
            meta = ext.notebook_metadata
            if meta is None and error:
                raise RuntimeError("No Notebook metadata available")
            elif meta is None:
                return None

            return self._parse_from_proxy(meta, error)

    def annotate(self, revision, sync=True):
        if revision.metadata is None:
            revision.metadata = {}

        try:
            if NOTEBOOK_NAME not in revision.metadata or NOTEBOOK_PATH not in revision.metadata:
                revision.metadata.update(self.parse_metadata())

            full_path = revision.metadata.get(NOTEBOOK_PATH)
            if sync and full_path and os.path.exists(full_path):
                revision.client.sync.sync(full_path, quiet=True)

        except Exception as e:  # pylint: disable=broad-exception-caught
            print(f"GoFigr could not automatically obtain the name of the currently"
                  f" running notebook. {PATH_WARNING} Details: {e}",
                  file=sys.stderr)

        return revision


class NotebookNameAnnotator(NotebookMetadataAnnotator):
    """(Deprecated) Annotates revisions with notebook name & path"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        print("NotebookNameAnnotator is deprecated. Please use NotebookMetadataAnnotator", file=sys.stderr)


class EnvironmentAnnotator(Annotator):
    """Annotates revisions with the python version & the kernel info"""
    def annotate(self, revision):
        if revision.metadata is None:
            revision.metadata = {}

        revision.metadata[NOTEBOOK_KERNEL] = sys.executable
        revision.metadata[PYTHON_VERSION] = sys.version

        return revision


class BackendAnnotator(Annotator):
    """Annotates revisions with the python version & the kernel info"""
    def annotate(self, revision):
        if revision.metadata is None:
            revision.metadata = {}

        revision.metadata[BACKEND_NAME] = revision.backend.get_backend_name() if revision.backend is not None else "N/A"
        return revision


class HistoryAnnotator(Annotator):
    """Annotates revisions with IPython execution history"""
    def annotate(self, revision):
        ip = get_ipython()

        if not hasattr(ip, 'history_manager'):
            return revision

        hist = ip.history_manager
        if hist is None:
            return revision

        revision.data.append(revision.client.CodeData(name="IPython history",
                                                      language="python",
                                                      format="jupyter-history/json",
                                                      contents=json.dumps(hist.input_hist_raw)))
        return revision



"""\
Copyright (c) 2022-2025, Flagstaff Solutions, LLC
All rights reserved.

"""

# pylint: disable=wildcard-import, unused-wildcard-import, undefined-variable

from gofigr.jupyter import *
get_ipython().run_line_magic('load_ext', 'gofigr')
configure(analysis=NotebookName())



"""\
Copyright (c) 2025, Flagstaff Solutions, LLC
All rights reserved.

"""
import json
import sys

from IPython import get_ipython

try:
    from databricks.sdk import WorkspaceClient
except ImportError:
    pass

def get_dbutils(shell=None):
    """Gets dbutils if running on DataBricks"""
    if shell is None:
        shell = get_ipython()

    return shell.user_ns.get('dbutils') if shell is not None else None

def is_databricks_environment():
    """True if running on Databricks"""
    return get_dbutils() is not None

def get_workspace_client():
    """Gets a WorkspaceClient if running on Databricks"""
    dbutils = get_dbutils()
    if dbutils is None:
        raise ValueError("dbutils not available. Are you sure you are running on Databricks?")

    databricks_url = dbutils.notebook.entry_point.getDbutils().notebook().getContext().apiUrl().getOrElse(None)
    token = dbutils.notebook.entry_point.getDbutils().notebook().getContext().apiToken().getOrElse(None)
    return WorkspaceClient(host=databricks_url, token=token)

def save_config(config):
    """Saves the GoFigr configuration in the databricks secret manager"""
    w = get_workspace_client()
    if not any(scope.name == "gofigr" for scope in w.secrets.list_scopes()):
        w.secrets.create_scope("gofigr")

    w.secrets.put_secret("gofigr", "config", string_value=json.dumps(config))

def get_config():
    """Gets the GoFigr config from the Databricks secret manager"""
    dbutils = get_dbutils()
    if dbutils is None:
        return None

    try:
        return json.loads(dbutils.secrets.get("gofigr", "config"))
    except Exception as e:  # pylint: disable=broad-exception-caught
        print(f"Failed to get config from DataBricks: {str(e)}", file=sys.stderr)
        return None



"""\
Implements a limited fully-offline mode for GoFigr.

Copyright (c) 2023-2025, Flagstaff Solutions, LLC
All rights reserved.

"""
# pylint: disable=use-dict-literal, ungrouped-imports

import asyncio
import io
import logging
import os
import sys
from collections import namedtuple
from datetime import timedelta, datetime

import PIL
import nest_asyncio
import numpy as np
from IPython import get_ipython
from ipywidgets import widgets
from ipywidgets.comm import create_comm

import plotly.graph_objects as go
from plotly.subplots import make_subplots

from gofigr import GoFigr
from gofigr.annotators import GitAnnotator, NotebookMetadataAnnotator, NOTEBOOK_NAME
from gofigr.backends.matplotlib import MatplotlibBackend
from gofigr.backends.plotly import PlotlyBackend
from gofigr.trap import SuppressDisplayTrap
from gofigr.utils import read_resource_b64, read_resource_binary
from gofigr.jupyter import _GoFigrExtension, _base_publish
from gofigr.publisher import _make_backend, _mark_as_published, \
    is_published, is_suppressed, suppress, infer_figure_and_backend
from gofigr.watermarks import DefaultWatermark, _qr_to_image, add_margins, stack_horizontally
from gofigr.widget import LiteStartupWidget

try:
    from IPython.core.display_functions import display
except ModuleNotFoundError:
    from IPython.core.display import display


PLOTNINE_PRESENT = False
try:
    import plotnine # pylint: disable=unused-import
    from gofigr.backends.plotnine import PlotnineBackend
    PLOTNINE_PRESENT = True
except ModuleNotFoundError:
    pass


logger = logging.getLogger(__name__)


EXTENSION_NAME = "_GF_EXTENSION_LITE"


def get_extension():
    """Returns the GoFigr Jupyter extension instance"""
    return get_ipython().user_ns.get(EXTENSION_NAME)


class _LiteGoFigrExtension(_GoFigrExtension):
    """Implementation of the GoFigr Jupyter extension for the "lite" (offline) backend."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, auto_publish=True, **kwargs)
        self.comm_data = None
        self.startup_widget = widgets.Output()
        self.publisher = LitePublisher()

    @property
    def is_ready(self):
        return True

    def post_run_cell(self, result):
        self.cell = result.info


def _gofigr_comm_handler(msg):
    ext = get_extension()
    first_data = ext.comm_data is None
    ext.comm_data = msg['content']['data']
    if first_data:
        with ext.startup_widget:
            LiteStartupWidget(get_extension()).show()


def _requeue(ip, digest, item):
    ip.kernel.session.digest_history = [x for x in ip.kernel.session.digest_history
                                        if x.decode('ascii') != digest.decode('ascii')]
    ip.kernel.msg_queue.put(item)


def load_ipython_extension(ip):
    """\
    Loads the "lite" Jupyter extension. The lite extension provides flexible
    figure watermarking but works offline and doesn't publish anything to
    GoFigr.io.

    :param ip: IPython shell
    :return: None

    """
    nest_asyncio.apply()

    ext = get_extension()

    if ext is not None:
        ext.unregister()

    ext = _LiteGoFigrExtension(ip)
    ext.register_hooks()
    ip.user_ns[EXTENSION_NAME] = ext
    ip.user_ns["get_extension"] = get_extension
    ip.user_ns["publish"] = publish
    ip.user_ns["suppress"] = suppress

    display(ext.startup_widget)

    my_comm = create_comm(target_name='gofigr', data={'state': "open"}, metadata={})
    my_comm.on_msg(_gofigr_comm_handler)

    # load_ipython_extension is running within the same event queue as Comms,
    # so the comm will not be processed until much later (in some cases this can be after the whole
    # notebook has finished executing (!)).
    #
    # That's too late to be useful, so we manually process the event queue message here.
    logger.debug("Waiting for comm message...")

    start_time = datetime.now()
    queue_copy = []
    comm_received = False
    while not comm_received and datetime.now() - start_time < timedelta(seconds=5):
        try:
            item = asyncio.run(ip.kernel.msg_queue.get(timeout=timedelta(seconds=1)))

            _, _, args = item
            _, msg = ip.kernel.session.feed_identities(*args, copy=False)
            digest = msg[0].bytes

            msg = ip.kernel.session.deserialize(msg, content=True, copy=False)

            if msg['msg_type'] == 'comm_msg' and msg['content']['comm_id'] == my_comm.comm_id:
                # This is our message!
                _gofigr_comm_handler(msg)
                comm_received = True

                # It's important to keep going until we exhaust all messages. Otherwise the previous messages
                # will be appended out-of-order.

            queue_copy.append((digest, item))

        except TimeoutError:
            if comm_received:  # no more messages and comm-received: we're done
                break
            else:
                logger.debug("Timeout. Waiting...")  # no more messages and still no comm. Wait.
                asyncio.run(asyncio.sleep(0.5))

    # Restore the queue
    for digest, item in queue_copy:
        _requeue(ip, digest, item)


class LiteClient(GoFigr):
    """
    A specialized implementation of the GoFigr class designed for local-only usage.

    LiteClient is a restricted version of the GoFigr client. It is intended for environments
    where local-only functionality is required, eliminating any backend interaction or
    synchronization capabilities. This class disables remote requests and syncing, ensuring
    that all functionality remains self-contained and isolated.

    :param url: Always set to None as LiteClient does not interact with a remote server.
    :param authenticate: Always set to False since authentication is not required
        for local-only operation.

    """

    def __init__(self, *args, **kwargs):
        super().__init__(url=None, authenticate=False, *args, **kwargs)

    def _unsupported(self):
        raise ValueError("GoFigr Lite is local-only.")

    def _request(self, *args, **kwargs):  # pylint: disable=unused-argument
        self._unsupported()

    @property
    def sync(self):
        raise ValueError("GoFigr Lite is local-only. Asset sync is not supported.")


class LitePlotlyBackend(PlotlyBackend):
    """Plotly backend for GoFigr"""
    def get_backend_name(self):
        return "plotly_lite"

    def add_interactive_watermark(self, fig, rev, watermark):
        # pylint: disable=too-many-locals

        # isinstance doesn't work if the extension is reloaded because it re-loads class definitions.
        # So we rely on duck typing as a workaround.
        if not hasattr(watermark, "get_table"):
            logger.debug(f"Skipping watermarking because it's not a LiteWatermark instance: {watermark}")
            return fig

        orig_height = getattr(fig.layout, "height")
        if orig_height is None:
            orig_height = 450

        orig_width = getattr(fig.layout, "width")
        if orig_width is None:
            orig_width = 700

        logo_width =  np.clip(0.1 * orig_width, 70, 200)

        logo_height = logo_width

        key_value_pairs = watermark.get_table(rev)
        table_height = 40 * (len(key_value_pairs) + 1)
        total_height = orig_height + table_height

        # Define the subplot grid with three distinct areas
        specs = [
            [{'type': 'xy', 'colspan': 2}, None],  # Top row for the main graph
            [{'type': 'xy'}, {'type': 'table'}]  # Bottom row for the image and table
        ]

        new_fig = make_subplots(
            rows=2,
            cols=2,
            shared_xaxes=False,
            vertical_spacing=0.1,
            row_heights=[orig_height / total_height, 1.0 - orig_height / total_height],
            column_widths=[logo_width / orig_width, 1.0 - logo_width / orig_width],  # Allocate more space to the table
            specs=specs
        )

        # Add all traces from the original figure to the top subplot
        for trace in fig.data:
            new_fig.add_trace(trace, row=1, col=1)

        # Add the GoFigr.io logo using add_layout_image
        logo_b64 = read_resource_b64("gofigr.resources", "logo_large.png")

        new_fig.add_layout_image(
            dict(
                source=f"data:image;base64,{logo_b64}",
                xref="paper",  # Anchor the image to the second subplot's x-axis
                yref="paper",  # Anchor the image to the second subplot's y-axis
                x=0.0,
                y=0.5 * (table_height - logo_height) / total_height,
                xanchor="left",
                yanchor="bottom",
                sizex=logo_width / orig_width,
                sizey=logo_height / orig_height,
                layer="below",
            )
        )

        # Make the subplot axes for the image invisible
        new_fig.update_xaxes(visible=False, row=2, col=1)
        new_fig.update_yaxes(visible=False, row=2, col=1)

        # Create the table
        header_values = ['Key', 'Value']
        cell_values = [
            list(key_value_pairs.keys()),
            list(key_value_pairs.values())
        ]

        table = go.Table(
            columnwidth=[100, 400],
            header=dict(values=header_values, align='left'),
            cells=dict(values=cell_values, align='left')
        )

        # Add the table to the bottom-right subplot
        new_fig.add_trace(table, row=2, col=2)

        # Update the overall layout
        new_fig.update_layout(
            title=fig.layout.title,
            height=orig_height + table_height
        )

        return new_fig


class LiteWatermark(DefaultWatermark):
    """
    Represents a specialized watermark generator extending DefaultWatermark.

    This class is designed to create lightweight watermarks for notebook figures
    with associated Git metadata and other contextual details. It provides custom
    methods to generate tables and optional QR codes with Git commit information.
    The generated watermark includes the user, notebook name, commit hash, and
    current date.

    :ivar show_qr_code: Flag indicating whether QR code generation is enabled.
    :type show_qr_code: bool
    :ivar qr_scale: Scale factor for QR code size.
    :type qr_scale: int
    :ivar qr_foreground: Foreground color of the QR code.
    :type qr_foreground: str
    :ivar qr_background: Background color of the QR code.
    :type qr_background: str
    :ivar margin_px: Pixel margin around the QR code.
    :type margin_px: int
    """

    def get_table(self, revision):
        """
        Retrieve a dictionary representing metadata information about a notebook revision.

        The method applies a series of annotators to enrich the given revision with
        additional metadata and then constructs a dictionary containing details such as
        the logged-in user, notebook name, commit details, and a timestamp of the operation.
        If certain metadata fields are not available, fallback default values are used.

        :param revision: The revision object whose metadata is to be processed.
            Must be annotated with `GitAnnotator` and `NotebookMetadataAnnotator`
            before metadata extraction.
        :type revision: Revision
        :return: A dictionary containing metadata details including user, notebook name,
            commit hash, timestamp, and a link to the Git commit. Defaults to placeholder
            values if the data is unavailable in the revision.
        :rtype: dict

        """
        GitAnnotator().annotate(revision)
        NotebookMetadataAnnotator().annotate(revision, sync=False)

        git_link = revision.metadata.get('git', {}).get('commit_link', None)
        git_hash = revision.metadata.get('git', {}).get('hash', None)
        git_link = git_link or "No git link available"

        notebook_name = revision.metadata.get(NOTEBOOK_NAME, "N/A")

        data = {"User": str(os.getlogin()),
               "Notebook": notebook_name,
               "Commit": git_hash,
               "Date": str(datetime.now()),
               "Git Link": git_link}
        return {k: v for k, v in data.items() if v is not None}

    def get_watermark(self, revision):
        """\
        Generates just the watermark for a revision.

        :param revision: FigureRevision
        :return: PIL.Image

        """
        data =  self.get_table(revision)
        table = [(f'{k}: ', v) for k, v in data.items() if k != "Git Link"]

        identifier_img = self.draw_table(table)
        git_link = data.get("Git Link", None)

        qr_img = None
        if self.show_qr_code and git_link:
            qr_img = _qr_to_image(git_link, scale=self.qr_scale,
                                  module_color=self.qr_foreground,
                                  background=self.qr_background)
            qr_img = add_margins(qr_img, self.margin_px)

        logo = PIL.Image.open(io.BytesIO(read_resource_binary("gofigr.resources", "logo_large.png")))
        logo_size = identifier_img.size[1]
        if qr_img and qr_img.size[1] > logo_size:
            logo_size = qr_img.size[1]

        logo.thumbnail((logo_size, logo_size))
        identifier_img = stack_horizontally(logo, identifier_img)

        return stack_horizontally(identifier_img, qr_img)


DEFAULT_LITE_BACKENDS = (MatplotlibBackend, LitePlotlyBackend)

if PLOTNINE_PRESENT:
    # pylint: disable=possibly-used-before-assignment
    DEFAULT_LITE_BACKENDS = (PlotnineBackend,) + DEFAULT_LITE_BACKENDS


_LoggedRevision = namedtuple("_LoggedRevision",
                             ["revision", "backend", "title", "displayed_image"])


class LitePublisher:
    """\
    "Lite" publisher: adds annotations to figures but doesn't publish anything to GoFigr.io.
    """
    # pylint: disable=too-many-arguments
    def __init__(self,
                 backends=None,
                 watermark=None,
                 show_watermark=True,
                 clear=True,
                 widget_class=None,
                 default_metadata=None,
                 log_revisions=False):
        self.gf = LiteClient()
        self.watermark = watermark or LiteWatermark()
        self.show_watermark = show_watermark
        self.backends = [_make_backend(bck) for bck in (backends or DEFAULT_LITE_BACKENDS)]
        self.clear = clear
        self.widget_class = widget_class
        self.default_metadata = default_metadata or {}
        self.revision_log = []
        self.log_revisions = log_revisions

    def auto_publish_hook(self, extension, data, suppress_display=None):
        """\
        Hook for automatically annotating figures.

        :param extension: GoFigrExtension instance
        :param data: Dictionary of mime formats.
        :param suppress_display: if used in an auto-publish hook, this will contain a callable which will
               suppress the display of this figure using the native IPython backend.

        :return: None
        """
        logger.debug("Running auto-publish hook")
        for backend in self.backends:
            logger.debug(f"Trying backend {backend}")
            compatible_figures = list(backend.find_figures(extension.shell, data))
            logger.debug(f"Found {len(compatible_figures)} figures")
            for fig in compatible_figures:
                logger.debug(f"Checking figure {fig}. Published: {is_published(fig)}. Suppressed: {is_suppressed(fig)}")
                if not is_published(fig) and not is_suppressed(fig):
                    logger.debug("Publishing...")
                    self.publish(fig=fig, backend=backend, suppress_display=suppress_display)

            if len(compatible_figures) > 0:
                break

    def _apply_watermark(self, fig, rev, backend, image_options):
        """\
        Extracts ImageData in various formats.

        :param backend: backend to use
        :param fig: figure object
        :param rev: LiteRevision object
        :param image_options: backend-specific parameters
        :return: tuple of: list of ImageData objects, watermarked image to display

        """
        if image_options is None:
            image_options = {}

        if backend.is_interactive(fig):
            return backend.add_interactive_watermark(fig, rev, self.watermark)
        else:
            img = PIL.Image.open(io.BytesIO(backend.figure_to_bytes(fig, "png", image_options)))
            img.load()
            return self.watermark.apply(img, rev)

    def publish(self, fig=None, metadata=None,
                backend=None, image_options=None,
                suppress_display=None, **kwargs):
        """\
        Publishes a revision to the server.

        :param fig: figure to publish. If None, we'll use plt.gcf()
        :param metadata: metadata (JSON) to attach to this revision
        :param backend: backend to use, e.g. MatplotlibBackend. If None it will be inferred automatically based on \
               figure type
        :param image_options: backend-specific params passed to backend.figure_to_bytes
        :param suppress_display: if used in an auto-publish hook, this will contain a callable which will
               suppress the display of this figure using the native IPython backend.

        :return: FigureRevision instance

        """
        if len(kwargs) > 0:
            print(f"GoFigr Lite does not support the following arguments: {', '.join(kwargs)}", file=sys.stderr)

        # pylint: disable=too-many-branches, too-many-locals
        fig, backend = infer_figure_and_backend(fig, backend, self.backends)

        combined_meta = dict(self.default_metadata) if self.default_metadata is not None else {}
        if metadata is not None:
            combined_meta.update(metadata)

        rev = self.gf.Revision(figure=None, metadata=combined_meta, backend=backend)  # pylint: disable=no-member

        logger.debug("Applying watermark to figure %s", fig)
        image_to_display = self._apply_watermark(fig, rev, backend, image_options)
        logger.debug("Watermark applied: %s", image_to_display)

        if image_to_display is not None:
            with SuppressDisplayTrap():
                logger.debug("Displaying watermarked figure")
                display(image_to_display)

            if suppress_display is not None:
                suppress_display()

        _mark_as_published(fig)

        if self.clear and self.show_watermark:
            backend.close(fig)

        if self.widget_class is not None:
            self.widget_class(rev).show()

        if self.log_revisions:
            self.revision_log.append(_LoggedRevision(rev, backend, backend.get_title(fig), image_to_display))

        return rev


def publish(fig=None, backend=None, **kwargs):
    """\
    Publishes a figure. See :func:`gofigr.lite.LitePublisher.publish` for a list of arguments. If figure and backend
    are both None, will publish default figures across all available backends.

    :param fig: figure to publish
    :param backend: backend to use
    :param kwargs:
    :return:

    """
    return _base_publish(ext=get_extension(), fig=fig, backend=backend, **kwargs)



"""\
Copyright (c) 2023-2025, Flagstaff Solutions, LLC
All rights reserved.

"""
# pylint: disable=protected-access, ungrouped-imports

import io
import os
import pickle
import sys

import PIL
try:
    from IPython.core.display_functions import display
except ModuleNotFoundError:
    from IPython.core.display import display

from gofigr import GoFigr, MeasureExecution, NotebookName
from gofigr.annotators import CellIdAnnotator, SystemAnnotator, CellCodeAnnotator, \
    PipFreezeAnnotator, NotebookMetadataAnnotator, EnvironmentAnnotator, BackendAnnotator, HistoryAnnotator, \
    GitAnnotator, Annotator, ScriptAnnotator
from gofigr.backends import get_backend, GoFigrBackend
from gofigr.backends.matplotlib import MatplotlibBackend
from gofigr.backends.plotly import PlotlyBackend
from gofigr.watermarks import DefaultWatermark
from gofigr.widget import DetailedWidget

PY3DMOL_PRESENT = False
if sys.version_info >= (3, 8):
    try:
        import py3Dmol  # pylint: disable=unused-import
        from gofigr.backends.py3dmol import Py3DmolBackend
        PY3DMOL_PRESENT = True
    except ModuleNotFoundError:
        pass

PLOTNINE_PRESENT = False
try:
    import plotnine # pylint: disable=unused-import
    from gofigr.backends.plotnine import PlotnineBackend
    PLOTNINE_PRESENT = True
except ModuleNotFoundError:
    pass

DEFAULT_ANNOTATORS = (NotebookMetadataAnnotator, EnvironmentAnnotator, CellIdAnnotator, CellCodeAnnotator,
                      SystemAnnotator, PipFreezeAnnotator, BackendAnnotator, HistoryAnnotator,
                      GitAnnotator, ScriptAnnotator)
DEFAULT_BACKENDS = (MatplotlibBackend, PlotlyBackend)
if PY3DMOL_PRESENT:
    # pylint: disable=possibly-used-before-assignment
    DEFAULT_BACKENDS = DEFAULT_BACKENDS + (Py3DmolBackend,)

if PLOTNINE_PRESENT:
    # pylint: disable=possibly-used-before-assignment
    DEFAULT_BACKENDS = (PlotnineBackend,) + DEFAULT_BACKENDS


def _mark_as_published(fig):
    """Marks the figure as published so that it won't be re-published again."""
    fig._gf_is_published = True
    return fig


def suppress(fig):
    """Suppresses the figure from being auto-published. You can still publish it by calling publish()."""
    fig._gf_is_suppressed = True
    return fig


def is_suppressed(fig):
    """Determines if the figure is suppressed from publication"""
    return getattr(fig, "_gf_is_suppressed", False)


def is_published(fig):
    """Returns True iff the figure has already been published"""
    return getattr(fig, "_gf_is_published", False)


def infer_figure_and_backend(fig, backend, enabled_backends):
    """\
    Given a figure and a backend where one of the values could be null, returns a complete set
    of a figure to publish and a matching backend.

    :param fig: figure to publish. None to publish the default for the backend
    :param backend: backend to use. If None, will infer from figure
    :param enabled_backends: list of enabled backends to use when inferring the backend
    :return: tuple of figure and backend
    """
    if fig is None and backend is None:
        raise ValueError("You did not specify a figure to publish.")
    elif fig is not None and backend is not None:
        return fig, backend
    elif fig is None and backend is not None:
        fig = backend.get_default_figure()

        if fig is None:
            raise ValueError("You did not specify a figure to publish, and the backend does not have "
                             "a default.")
    else:
        backend = get_backend(fig, enabled_backends)

    return fig, backend


# pylint: disable=too-many-instance-attributes
class Publisher:
    """\
    Publishes revisions to the GoFigr server.
    """
    # pylint: disable=too-many-arguments
    def __init__(self,
                 gf=None,
                 workspace=None,
                 analysis=None,
                 annotators=None,
                 backends=None,
                 watermark=None,
                 show_watermark=True,
                 image_formats=("png", "eps", "svg"),
                 interactive=True,
                 default_metadata=None,
                 clear=False,
                 save_pickle=True,
                 widget_class=DetailedWidget):
        """

        :param gf: GoFigr instance
        :param annotators: revision annotators
        :param backends: figure backends, e.g. MatplotlibBackend
        :param watermark: watermark generator, e.g. QRWatermark()
        :param show_watermark: True to show watermarked figures instead of original.
               False to always display the unmodified figure. Default True.
        :param image_formats: image formats to save by default
        :param interactive: whether to publish figure HTML if available
        :param clear: whether to close the original figures after publication
        :param save_pickle: if True, will save the figure in pickle format in addition to any of the image formats
        :param widget_class: Widget type to show, e.g. DetailedWidget or CompactWidget. It will appear below the
               published figure

        """
        self.gf = gf or GoFigr()
        self.watermark = watermark or DefaultWatermark()
        self.show_watermark = show_watermark
        self.annotators = [_make_annotator(ann) for ann in (annotators or DEFAULT_ANNOTATORS)]
        self.backends = [_make_backend(bck) for bck in (backends or DEFAULT_BACKENDS)]
        self.image_formats = image_formats
        self.interactive = interactive
        self.clear = clear
        self.default_metadata = default_metadata
        self.save_pickle = save_pickle
        self.widget_class = widget_class

        self.workspace = self.gf.find_workspace(workspace).fetch()
        self.analysis = self.gf.find_analysis(self.workspace, analysis)

    def _check_analysis(self):
        if self.analysis is None:
            print("You did not specify an analysis to publish under. Please call "
                  "configure(...) and specify one. See "
                  "https://gofigr.io/docs/gofigr-python/latest/gofigr.html#gofigr.jupyter.configure.",
                  file=sys.stderr)
            return None
        elif isinstance(self.analysis, NotebookName):
            print("Your analysis is set to the name of this notebook, but the name could "
                  "not be inferred. Please call "
                  "configure(...) and specify the analysis manually. See "
                  "https://gofigr.io/docs/gofigr-python/latest/gofigr.html#gofigr.jupyter.configure.",
                  file=sys.stderr)
            return None
        else:
            return self.analysis

    def _resolve_target(self, fig, target, backend):
        analysis = self._check_analysis()
        if analysis is None:
            return None

        if target is None:
            # Try to get the figure's title
            fig_name = backend.get_title(fig)
            if fig_name is None:
                print("Your figure doesn't have a title and will be published as 'Anonymous Figure'. "
                      "To avoid this warning, set a figure title or manually call publish() with a target figure. "
                      "See https://gofigr.io/docs/gofigr-python/latest/start.html#publishing-your-first-figure for "
                      "an example.", file=sys.stderr)
                fig_name = "Anonymous Figure"

            sys.stdout.flush()
            return analysis.get_figure(fig_name, create=True)
        else:
            return self.gf.find_figure(analysis, target)

    def _get_pickle_data(self, gf, fig, revision, target):
        if not self.save_pickle:
            return []

        try:
            bio = io.BytesIO()
            pickle.dump(fig, bio)
            bio.seek(0)

            return [gf.FileData(name=f"{target.name}_rev_{revision.revision_index + 1}.pickle",
                                 data=bio.getvalue())]
        except Exception as e: # pylint: disable=broad-exception-caught
            print(f"WARNING: We could not obtain the figure in pickle format: {e}", file=sys.stderr)
            return []

    def _get_image_data(self, gf, backend, fig, rev, target, image_options):
        """\
        Extracts ImageData in various formats.

        :param gf: GoFigr instance
        :param backend: backend to use
        :param fig: figure object
        :param rev: gf.Revision object
        :param target: gf.Figure object
        :param image_options: backend-specific parameters
        :return: tuple of: list of ImageData objects, watermarked image to display

        """
        # pylint: disable=too-many-locals

        if image_options is None:
            image_options = {}

        image_to_display = None
        image_data = []
        for fmt in self.image_formats:
            if fmt.lower() not in backend.get_supported_image_formats():
                continue

            if fmt.lower() == "png":
                img = PIL.Image.open(io.BytesIO(backend.figure_to_bytes(fig, fmt, image_options)))
                img.load()
                watermarked_img = self.watermark.apply(img, rev)
            else:
                watermarked_img = None

            # First, save the image without the watermark
            try:
                image_data.append(gf.ImageData(name="figure",
                                               format=fmt,
                                               data=backend.figure_to_bytes(fig, fmt, image_options),
                                               is_watermarked=False))
            except Exception as e:  # pylint: disable=broad-exception-caught
                print(f"WARNING: We could not obtain the figure in {fmt.upper()} format: {e}", file=sys.stderr)
                continue

            # Now, save the watermarked version (if available)
            if watermarked_img is not None:
                bio = io.BytesIO()
                watermarked_img.save(bio, format=fmt)
                img_data = gf.ImageData(name="figure", format=fmt, data=bio.getvalue(),
                                        is_watermarked=True)
                image_data.append(img_data)

                if fmt.lower() == 'png':
                    image_to_display = img_data

        if self.interactive and backend.is_interactive(fig):
            image_data.append(gf.ImageData(name="figure", format="html",
                                           data=backend.figure_to_html(fig).encode('utf-8'),
                                           is_watermarked=False))

            wfig = backend.add_interactive_watermark(fig, rev, self.watermark)
            html_with_watermark = gf.ImageData(name="figure", format="html",
                                               data=backend.figure_to_html(wfig).encode('utf-8'),
                                               is_watermarked=True)
            image_data.append(html_with_watermark)
            image_to_display = wfig  # display the native Figure

        image_data.extend(self._get_pickle_data(gf, fig, rev, target))

        return image_data, image_to_display

    def annotate(self, rev, annotators=None):
        """
        Annotates a FigureRevision using self.annotators.
        :param rev: revision to annotate
        :param annotators: list of annotators to use. Defaults to self.annotators
        :return: annotated revision

        """
        for annotator in (annotators or self.annotators):
            with MeasureExecution(annotator.__class__.__name__):
                annotator.annotate(rev)

        # Add data annotations
        rev.assets = [self.gf.AssetLinkedToFigure(figure_revision=rev,
                                                  asset_revision=data_rev,
                                                  use_type="indirect") for data_rev in self.gf.sync.revisions]
        return rev

    def _prepare_files(self, gf, files):
        if not isinstance(files, dict):
            files = {os.path.basename(p): p for p in files}

        data = []
        for name, filelike in files.items():
            if isinstance(filelike, str): # path
                with open(filelike, 'rb') as f:
                    data.append(gf.FileData(data=f.read(), name=name, path=filelike))
            else:  # stream
                data.append(gf.FileData(data=filelike.read(), name=name, path=None))

        return data

    def publish(self, fig=None, target=None, dataframes=None, metadata=None,
                backend=None, image_options=None, suppress_display=None, files=None,
                annotators=None):
        """\
        Publishes a revision to the server.

        :param fig: figure to publish. If None, we'll use plt.gcf()
        :param target: Target figure to publish this revision under. Can be a gf.Figure instance, an API ID, \
               or a FindByName instance.
        :param dataframes: dictionary of dataframes to associate & publish with the figure
        :param metadata: metadata (JSON) to attach to this revision
        :param backend: backend to use, e.g. MatplotlibBackend. If None it will be inferred automatically based on \
               figure type
        :param image_options: backend-specific params passed to backend.figure_to_bytes
        :param suppress_display: if used in an auto-publish hook, this will contain a callable which will
               suppress the display of this figure using the native IPython backend.
        :param files: either (a) list of file paths or (b) dictionary of name to file path/file obj
        :param annotators: list of annotators to use. Defaults to self.annotators

        :return: FigureRevision instance

        """
        # pylint: disable=too-many-branches, too-many-locals
        fig, backend = infer_figure_and_backend(fig, backend, self.backends)

        with MeasureExecution("Resolve target"):
            target = self._resolve_target(fig, target, backend)
            if getattr(target, 'revisions', None) is None:
                target.fetch()

        combined_meta = dict(self.default_metadata) if self.default_metadata is not None else {}
        if metadata is not None:
            combined_meta.update(metadata)

        with MeasureExecution("Bare revision"):
            # Create a bare revision first to get the API ID
            rev = self.gf.Revision(figure=target, metadata=combined_meta, backend=backend)
            target.revisions.create(rev)

        with MeasureExecution("Annotators"):
            # Annotate the revision
            self.annotate(rev, annotators=annotators)

        with MeasureExecution("Image data"):
            rev.image_data, image_to_display = self._get_image_data(self.gf, backend, fig, rev, target, image_options)

        if image_to_display is not None and self.show_watermark:
            if isinstance(image_to_display, self.gf.ImageData):
                display(image_to_display.image)
            else:
                display(image_to_display)

            if suppress_display is not None:
                suppress_display()

        if dataframes is not None:
            table_data = []
            for name, frame in dataframes.items():
                table_data.append(self.gf.TableData(name=name, dataframe=frame))

            rev.table_data = table_data

        if files is not None:
            rev.file_data = self._prepare_files(self.gf, files)

        with MeasureExecution("Final save"):
            rev.save(silent=True)

            # Calling .save() above will update internal properties based on the response from the server.
            # In our case, this will result in rev.figure becoming a shallow object with just the API ID. Here
            # we restore it from our cached copy, to avoid a separate API call.
            rev.figure = target

        _mark_as_published(fig)

        if self.clear and self.show_watermark:
            backend.close(fig)

        self.widget_class(rev).show()
        return rev


def _make_backend(backend):
    if isinstance(backend, GoFigrBackend):
        return backend
    else:
        return backend()


def _make_annotator(annotator):
    if isinstance(annotator, Annotator):
        return annotator
    else:
        return annotator()



"""\
Copyright (c) 2023, Flagstaff Solutions, LLC
All rights reserved.

"""
import io
import os
import re
import subprocess
import sys
import time
import traceback
from argparse import ArgumentParser
from datetime import datetime

import selenium
from selenium import webdriver
from selenium.webdriver.chrome.service import Service as ChromeService
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.support import wait, expected_conditions
from selenium.webdriver.support.wait import WebDriverWait
from webdriver_manager.chrome import ChromeDriverManager
from selenium.webdriver.common.by import By

def find_element_with_alternatives(driver, by, possible_values, delay_seconds=0.5, max_wait_seconds=30):
    """Calls driver.find_element using alternative names until the element is found, or raises an exception"""
    print(f"Looking for {possible_values}")

    start_time = datetime.now()
    while (datetime.now() - start_time).total_seconds() < max_wait_seconds:
        for val in possible_values:
            try:
                elt = driver.find_element(by=by, value=val)
                print(f"  => Found {val}")
                return elt
            except selenium.common.exceptions.NoSuchElementException:
                time.sleep(delay_seconds)
                continue  # try next possible value

    raise RuntimeError(f"No such element. Tried: " + ", ".join(possible_values))


def run_notebook(driver, jupyter_url, notebook_path):
    """Runs a notebook in the classic notebook UI"""
    print("Running classic notebook...")

    if '/tree' in jupyter_url:
        # Running Notebook 7
        nav_url = jupyter_url.replace("/tree?token=", f"/notebooks/{notebook_path}?factory=Notebook&token=")
    else:
        nav_url = jupyter_url.replace("?token=", f"notebooks/{notebook_path}?token=")

    driver.get(nav_url)
    print(f"Navigating to {nav_url}...")

    WebDriverWait(driver, 120).until(expected_conditions.visibility_of_element_located((By.ID,
                                             "Integration-tests-for-the-GoFigr-Python-client")))

    try:
        # For Jupyter Notebook 6.x
        find_element_with_alternatives(driver, by=By.CSS_SELECTOR, possible_values=["#kernellink"]).click()
        find_element_with_alternatives(driver, by=By.CSS_SELECTOR, possible_values=["#restart_run_all"]).click()
    except RuntimeError:
        # For Jupyter Notebook 5.x or 7.0+
        find_element_with_alternatives(driver, by=By.CSS_SELECTOR,
                                       possible_values=[
                                           'button[data-jupyter-action="jupyter-notebook:confirm-restart-kernel-and-run-all-cells"]',
                                           "button[data-command='runmenu:restart-and-run-all']",
                                           "button[data-command='notebook:restart-run-all']",
                                           "jp-button[data-command='runmenu:restart-and-run-all']",
                                           "jp-button[data-command='notebook:restart-run-all']",
                                       ]).click()

    # Confirm
    print("Confirming...")
    try:
        find_element_with_alternatives(driver, by=By.CSS_SELECTOR,
                                       possible_values=[".modal-dialog button.btn.btn-danger",
                                                        ".jp-Dialog-button.jp-mod-warn"]).click()
    except RuntimeError:
        # For some reason Jupyter doesn't prompt for confirmation on Linux
        driver.save_screenshot("confirmation_failed.png")
        print("Confirmation failed, but continuing anyway...")

    print("UI done. Waiting for execution...")


def run_lab(driver, jupyter_url, notebook_path):
    driver.get(jupyter_url.replace("/lab?token=", f"/lab/tree/{notebook_path}?token="))

    WebDriverWait(driver, 120).until(expected_conditions.visibility_of_element_located((By.CSS_SELECTOR,
                                             '[data-jupyter-id="Integration-tests-for-the-GoFigr-Python-client"]')))

    # Restart and run all button
    find_element_with_alternatives(driver, by=By.CSS_SELECTOR, possible_values=[
        "button[data-command='runmenu:restart-and-run-all']",
        "button[data-command='notebook:restart-run-all']",
        "jp-button[data-command='runmenu:restart-and-run-all']",
        "jp-button[data-command='notebook:restart-run-all']"
    ]).click()

    # Confirm
    find_element_with_alternatives(driver, by=By.CSS_SELECTOR, possible_values=[
        ".jp-Dialog-button.jp-mod-warn"
    ]).click()


def run_attempt(args, working_dir, reader, writer, attempt):
    proc = subprocess.Popen(["jupyter", args.service, "--no-browser", args.notebook_path],
                            stdout=writer,
                            stderr=writer)

    start_time = datetime.now()
    timed_out = True

    driver = None
    success = False
    try:
        jupyter_url = None
        while proc.poll() is None and jupyter_url is None:
            line = reader.readline()
            m = re.match(r'.*(http.*\?token=\w+).*', line)
            if m is not None:
                jupyter_url = m.group(1)
            elif "/tree" in line:
                raise RuntimeError("Found a URL but not a token. Are you using password authentication?")

            if (datetime.now() - start_time).total_seconds() >= args.timeout:
                raise RuntimeError("Timed out")

            time.sleep(0.5)

        output_path = os.path.join(working_dir, "integration_test.json")
        if os.path.exists(output_path):
            os.remove(output_path)
            print(f"Deleted {output_path}")

        if jupyter_url is None:
            raise RuntimeError("Jupyter URL unavailable. Did it start correctly?")

        print(f"URL: {jupyter_url}")
        time.sleep(2)

        print("Starting Chrome...")
        opts = Options()
        opts.add_argument('--window-size=1920,10000')
        if args.headless:
            opts.add_argument('--headless=new')

        print(f"Headless: {args.headless}")
        driver = webdriver.Chrome(service=ChromeService(ChromeDriverManager().install()),
                                  options=opts)
        driver.implicitly_wait(5.0)

        if args.service == "notebook":
            run_notebook(driver, jupyter_url, args.notebook_path)
        elif args.service == "lab":
            run_lab(driver, jupyter_url, args.notebook_path)
        else:
            raise ValueError(f"Unsupported service: {args.service}")

        while (datetime.now() - start_time).total_seconds() < args.timeout:
            if os.path.exists(output_path + ".done"):
                timed_out = False
                success = True
                break

            time.sleep(1)

        if timed_out:
            print("Execution timed out.", file=sys.stderr)
    except:
        traceback.print_exc()
        print("Execution failed", file=sys.stderr)
    finally:
        if driver is not None:
            driver.save_screenshot(os.path.join(working_dir, f"screenshot_attempt{attempt}.png"))

            with open(os.path.join(working_dir, f"attempt{attempt}.html"), 'w') as f:
                f.write(driver.page_source)

            driver.close()
            time.sleep(5)

        proc.terminate()
        time.sleep(5)

        return success


def main():
    parser = ArgumentParser(description="Uses Selenium to run a Jupyter notebook inside a Notebook/Lab server"
                                        " instance.")
    parser.add_argument("service", help="notebook or lab")
    parser.add_argument("notebook_path", help="Path to ipynb notebook")
    parser.add_argument("--timeout", type=int, default=60*15,
                        help="Timeout in seconds for the notebook to finish execution")
    parser.add_argument("--headless", action="store_true", help="Run in headless mode")
    parser.add_argument("--retries", type=int, default=2, help="Maximum number of execution attempts.")
    args = parser.parse_args()

    working_dir = os.path.dirname(args.notebook_path)
    attempt = 0
    success = False
    while attempt < args.retries and not success:
        print(f"Running attempt {attempt + 1}...")
        filename = os.path.join(working_dir, f"jupyter_attempt{attempt + 1}.log")

        with io.open(filename, "w") as writer, \
                io.open(filename, "r", 1) as reader:
            success = run_attempt(args, working_dir, reader, writer, attempt + 1)
            attempt += 1

    status = "Succeeded" if success else "Failed"
    print(f"{status} after {attempt} attempts.")

    sys.exit(0 if success else 1)


if __name__ == "__main__":
    main()



"""\
Copyright (c) 2023, Flagstaff Solutions, LLC
All rights reserved.

"""
import concurrent
import datetime
import json
import os
import shutil
import subprocess
import sys
from argparse import ArgumentParser

from teamcity.messages import TeamcityServiceMessages

from gofigr import GoFigr, find_config, FindByName
from gofigr.utils import from_config_or_env


@from_config_or_env("GF_", find_config())
def get_gf(test_user, test_password):
    gf = GoFigr(username=test_user, password=test_password, api_key=None, url="https://api-dev.gofigr.io")
    if "api-dev" not in gf.api_url:
        raise RuntimeError("Running against production. Stopping.")
    return gf


def clean_up(analysis=None, clean_assets=False):
    gf = get_gf()

    worx = gf.find_workspace(FindByName("Integration tests", create=True))
    worx.fetch()

    if analysis:
        analyses_to_clean = [analysis]
    else:
        analyses_to_clean = worx.analyses

    print("Cleaning up....")
    for ana in analyses_to_clean:
        ana.fetch()

        for fig in ana.figures:
            fig.fetch()
            for rev in fig.revisions:
                rev.delete(delete=True)

            fig.delete(delete=True)

    if clean_assets:
        for asset in worx.assets:
            asset.delete(delete=True)

    print("Cleanup complete.")


def format_cmd(args):
    """Formats command line arguments."""
    def format_one(x):
        if ' ' in x:
            return f'"{x}"'
        else:
            return x

    return " ".join([format_one(x) for x in args])


def run_one_config(args, idx, config, all_configurations, messages):
    """Runs a single configuration"""
    run_one = os.path.join(os.path.dirname(sys.argv[0]), "run_one.sh")
    out_dir = os.path.join(args.output, config["name"])

    gf = get_gf()
    worx = gf.find_workspace(FindByName("Integration tests", create=True))
    ana = worx.get_analysis(config["name"])

    clean_up(analysis=ana)

    notebook = config.get('notebook', 'integration_tests.ipynb')
    results_file = config.get('results_file', 'integration_test.json')

    print(f"Running configuration {idx + 1}/{len(all_configurations)}: ")
    print(f'  * Name: {config["name"]}')
    print(f"  * Python: {config['python']}")
    print(f"  * Notebook: {notebook}")
    print(f"  * Result file: {results_file}")
    print(f"  * Dependencies: {config['dependencies']}")
    print(f"  * Directory: {out_dir}")

    if os.path.exists(out_dir):
        if args.force:
            shutil.rmtree(out_dir)
        else:
            print("  => Path exists. Skipping\n")
            messages.testIgnored(config["name"], flowId=config["name"])
            return True

    os.makedirs(out_dir, exist_ok=True)

    with open(os.path.join(out_dir, "config.json"), 'w') as f:
        json.dump(config, f)

    with open(os.path.join(out_dir, ".gofigr"), 'w') as f:
        json.dump({
            "username": os.environ["GF_TEST_USER"],
            "password": os.environ["GF_TEST_PASSWORD"],
            "workspace": worx.api_id,
            "analysis": ana.api_id,
            "url": "https://api-dev.gofigr.io",
            "auto_publish": True
        }, f)

    runner_args = ["bash", run_one, out_dir, os.path.join(out_dir, "driver.log"),
                   config["python"], config["service"],
                   config["dependencies"],
                   notebook, results_file]
    if not args.no_headless:
        runner_args.append("--headless")

    print(f"  => Running {format_cmd(runner_args)}")
    cp = subprocess.run(runner_args, check=False)
    print(f"Exit code: {cp.returncode}")
    if cp.returncode != 0:
        print(f"  => Process failed with code {cp.returncode}")

        with open(os.path.join(out_dir, "errors.json"), 'w') as ef:
            json.dump({"error": None if cp.stderr is None else cp.stderr.decode('ascii', errors='ignore')}, ef)

        return {'success': False, 'analysis': ana, 'log_path': os.path.join(out_dir, "driver.log")}
    else:
        print("  => Complete")
        return {'success': True, 'analysis': ana}




def test_wrapper(messages, config, args, idx, all_configurations):
    test_name = config["name"].replace(".", "_")
    
    start_time = datetime.datetime.now()
    analysis = None
    try:
        messages.testStarted(test_name, flowId=test_name)
        res = run_one_config(args, idx, config, all_configurations, messages)

        analysis = res['analysis']
        if not res['success']:
            with open(res['log_path'], 'r') as f:
                messages.testFailed(test_name, flowId=test_name, message="Test failed", details=f.read())
        else:
            messages.testFinished(test_name, testDuration=datetime.datetime.now() - start_time,
                                  flowId=test_name)
    except Exception as e:  # pylint: disable=broad-exception-caught
        messages.testFailed(test_name, message=str(e), flowId=test_name)
    finally:
        if analysis is not None and not args.no_cleanup:
            clean_up(analysis=analysis)


def main():
    parser = ArgumentParser(description="Runs integration tests based on a config file")
    parser.add_argument("config", help="config file (JSON)")
    parser.add_argument("output", help="output directory")
    parser.add_argument("--no-headless", action="store_true", help="Whether to NOT run in headless mode. Default is true.")
    parser.add_argument("--force", action="store_true", help="Force re-run even if directory already exists")
    parser.add_argument("--threads", action="store", type=int, help="Number of threads to use", default=10)
    parser.add_argument("--subset", action="store", nargs="+", help="Specify names of configurations to run")
    parser.add_argument("--no-cleanup", action="store_true", help="Do not clean up after running")
    args = parser.parse_args()

    with open(args.config, "r") as f:
        all_configurations = json.load(f)

        if isinstance(all_configurations, dict):
            all_configurations = [all_configurations]

    if args.subset and len(args.subset) > 0:
        all_configurations = [conf for conf in all_configurations if conf["name"] in args.subset]

    messages = TeamcityServiceMessages()
    messages.testSuiteStarted("Compatibility checks")
    messages.testCount(len(all_configurations))

    clean_up(clean_assets=True)

    with concurrent.futures.ThreadPoolExecutor(max_workers=args.threads) as executor:
        for idx, config in enumerate(all_configurations):
            executor.submit(test_wrapper, messages, config, args, idx, all_configurations)

    if not args.no_cleanup:
        clean_up(analysis=None, clean_assets=True)

    messages.testSuiteFinished("Compatibility checks")

if __name__ == "__main__":
    main()



"""\
Copyright (c) 2024, Flagstaff Solutions, LLC
All rights reserved.

"""
from abc import ABC
# pylint: disable=trailing-whitespace, ungrouped-imports

from base64 import b64encode
from uuid import uuid4

import humanize
from IPython.core.display import HTML

from gofigr.annotators import NotebookMetadataAnnotator, NOTEBOOK_NAME

try:
    from IPython.core.display_functions import display
except ModuleNotFoundError:
    from IPython.core.display import display

from gofigr.utils import read_resource_b64, read_resource_text


BUTTON_STYLE = "padding-top: 0.25em; padding-bottom: 0.25em; padding-left: 0.5em; padding-right: 0.5em; " + \
               "color: #fff; display: inline-block; font-weight: 400; line-height: 1; " + \
               "text-align: center; vertical-align: middle; cursor: pointer; border: 1px solid transparent; " + \
               "font-size: 0.875rem; border-radius: 0.35rem; margin-top: auto; margin-bottom: auto; "

VIEW_BUTTON_STYLE = BUTTON_STYLE + "background-color: #ed6353; border-color: #ed635e; margin-left: 0.5rem; "
DOWNLOAD_BUTTON_STYLE = BUTTON_STYLE + "background-color: #0d53b1; border-color: #0d53b1; margin-left: 0.5rem; "

COPY_BUTTON_STYLE = BUTTON_STYLE + "background-color: #2c7df5; border-color: #2c7df5; margin-left: 0.5rem; "

WIDGET_STYLE = "margin-top: 1rem; margin-bottom: 1rem; margin-left: auto; margin-right: auto;" + \
               "display: flex !important; border: 1px solid transparent; border-color: #cedcef; padding: 0.75em; " + \
               "border-radius: 0.35rem; flex-wrap: wrap; "

COMPACT_WIDGET_STYLE = "margin-top: 1rem; margin-bottom: 1rem; margin-left: auto; margin-right: auto;" + \
               "display: flex !important; border: 1px solid transparent; border-color: #cedcef; padding: 0.25em; " + \
               "padding-top: 0.5em; " + \
               "border-radius: 0.35rem; flex-wrap: wrap; "

MINIMAL_WIDGET_STYLE = "margin-top: 1rem; margin-bottom: 1rem; margin-left: auto; margin-right: auto;" + \
               "display: flex !important; padding: 0.25em; " + \
               "padding-top: 0.5em; flex-wrap: wrap; "

ME2 = "margin-right: 1rem"

MESSAGE_STYLE = "padding-top: 0.25rem; "

ROW_STYLE = "width: 100%; display: flex; flex-wrap: wrap;"

BREAK_STYLE = "flex-basis: 100%; height: 0;"

FA_VIEW = """<svg xmlns="http://www.w3.org/2000/svg" height="0.85rem" width="0.85rem" viewBox="0 0 512 512"><!--!Font 
Awesome Free 6.5.1 by @fontawesome - https://fontawesome.com License - https://fontawesome.com/license/free Copyright 
2024 Fonticons, Inc.--><path fill="#ffffff" d="M352 0c-12.9 0-24.6 7.8-29.6 19.8s-2.2 25.7 6.9 34.9L370.7 96 201.4 
265.4c-12.5 12.5-12.5 32.8 0 45.3s32.8 12.5 45.3 0L416 141.3l41.4 41.4c9.2 9.2 22.9 11.9 34.9 6.9s19.8-16.6 
19.8-29.6V32c0-17.7-14.3-32-32-32H352zM80 32C35.8 32 0 67.8 0 112V432c0 44.2 35.8 80 80 80H400c44.2 0 80-35.8 
80-80V320c0-17.7-14.3-32-32-32s-32 14.3-32 32V432c0 8.8-7.2 16-16 16H80c-8.8 0-16-7.2-16-16V112c0-8.8 7.2-16 
16-16H192c17.7 0 32-14.3 32-32s-14.3-32-32-32H80z"/></svg>"""

FA_FEATHER = """<svg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 640 640">
<!--!Font Awesome Free v7.0.1 by @fontawesome - https://fontawesome.com License - 
https://fontawesome.com/license/free Copyright 2025 Fonticons, Inc.-->
<path d="M416 64C457 64 496.3 80.3 525.2 109.2L530.7 114.7C559.7 143.7 576 183 576 223.9C576 248 570.3 
271.5 559.8 292.7C557.9 296.4 554.5 299.2 550.5 300.4L438.5 334C434.6 335.2 432 338.7 432 342.8C432 347.9 436.1 352 
441.2 352L473.4 352C487.7 352 494.8 369.2 484.7 379.3L462.3 401.7C460.4 403.6 458.1 404.9 455.6 405.7L374.6 430C370.7 
431.2 368.1 434.7 368.1 438.8C368.1 443.9 372.2 448 377.3 448C390.5 448 396.2 463.7 385.1 470.9C344 497.5 295.8 512 
246.1 512L160.1 512L112.1 560C103.3 568.8 88.9 568.8 80.1 560C71.3 551.2 71.3 536.8 80.1 528L320 288C328.8 279.2 
328.8 264.8 320 256C311.2 247.2 296.8 247.2 288 256L143.5 400.5C137.8 406.2 128 402.2 128 394.1C128 326.2 155
 261.1 203 213.1L306.8 109.2C335.7 80.3 375 64 416 64z"/></svg>
"""

FA_COPY = """<svg xmlns="http://www.w3.org/2000/svg" height="1rem" width="0.85rem" viewBox="0 0 448 512"><!--!Font 
Awesome Free 6.5.1 by @fontawesome - https://fontawesome.com License - https://fontawesome.com/license/free Copyright 
2024 Fonticons, Inc.--><path fill="#ffffff" d="M208 0H332.1c12.7 0 24.9 5.1 33.9 14.1l67.9 67.9c9 9 14.1 21.2 14.1 
33.9V336c0 26.5-21.5 48-48 48H208c-26.5 0-48-21.5-48-48V48c0-26.5 21.5-48 48-48zM48 128h80v64H64V448H256V416h64v48c0 
26.5-21.5 48-48 48H48c-26.5 0-48-21.5-48-48V176c0-26.5 21.5-48 48-48z"/></svg>"""

FA_COPY_LIGHT = FA_COPY.replace("#ffffff", "#0d53b1")

FA_DOWNLOAD = """<svg xmlns="http://www.w3.org/2000/svg" height="1rem" width="1rem" viewBox="0 0 512 512"><!--!Font 
Awesome Free 6.5.1 by @fontawesome - https://fontawesome.com License - https://fontawesome.com/license/free Copyright 
2024 Fonticons, Inc.--><path fill="#ffffff" d="M288 32c0-17.7-14.3-32-32-32s-32 14.3-32 
32V274.7l-73.4-73.4c-12.5-12.5-32.8-12.5-45.3 0s-12.5 32.8 0 45.3l128 128c12.5 12.5 32.8 12.5 45.3 
0l128-128c12.5-12.5 12.5-32.8 0-45.3s-32.8-12.5-45.3 0L288 274.7V32zM64 352c-35.3 0-64 28.7-64 64v32c0 35.3 28.7 64 
64 64H448c35.3 0 64-28.7 64-64V416c0-35.3-28.7-64-64-64H346.5l-45.3 45.3c-25 25-65.5 25-90.5 0L165.5 352H64zm368 
56a24 24 0 1 1 0 48 24 24 0 1 1 0-48z"/></svg>"""


def timestamp_to_local_tz(dt_tz):
    """Converts a datetime to the local timezone"""
    return dt_tz.astimezone(tz=None)


class WidgetBase(ABC):
    """Generates HTML/Javascript for the GoFigr Jupyter widget shown under each figure"""

    def __init__(self):
        """

        :param revision: gf.Revision instance
        """
        self._logo_b64 = None

    def get_logo_b64(self):
        """Reads the GoFigr logo and returns it as a base64 string"""
        if self._logo_b64 is not None:
            return self._logo_b64

        self._logo_b64 = read_resource_b64("gofigr.resources", "logo_small.png")
        return self._logo_b64

    def show(self):
        """Displays the widget in Jupyter"""
        raise NotImplementedError


class RevisionWidgetBase(WidgetBase, ABC):
    """Generates HTML/Javascript for the GoFigr Jupyter widget shown under each figure"""

    def __init__(self, revision):
        """

        :param revision: gf.Revision instance
        """
        super().__init__()
        self.revision = revision
        self.id = f"gf-widget-{str(uuid4())}"
        self.alert_id = f"{self.id}-alert"

    def get_download_link(self, label, watermark, image_format="png"):
        """Generates HTML for the download link"""
        matches = [img for img in self.revision.image_data
                   if img.format and img.format.lower() == image_format.lower() and img.is_watermarked == watermark]
        if len(matches) == 0:
            return ""
        else:
            img = matches[0]
            return f"""<a download="{self.revision.figure.name}.{image_format}" 
                          style="{DOWNLOAD_BUTTON_STYLE}"
                          href="data:image/{image_format};base64,{b64encode(img.data).decode('ascii')}">
                          {label}</a>"""

    def _render_template(self, name, mapping):
        """Renders a JavaScript template with the given variable mappings"""
        template = read_resource_text("gofigr.resources", name).replace("\n", "")
        for key, value in mapping.items():
            template = template.replace(key, value)
        template = template.replace("\"", "'")
        return template

    def get_copy_link(self, label, watermark, image_format="png"):
        """Generates HTML/JS for the copy image button"""
        matches = [img for img in self.revision.image_data
                   if img.format and img.format.lower() == image_format.lower() and img.is_watermarked == watermark]
        if len(matches) == 0:
            return ""

        img = matches[0]
        onclick = self._render_template("copy_image.js",
                                        {"_ALERT_ID_": self.alert_id,
                                         "_ERROR_MESSAGE_": "Unable to copy image. "
                                                            "Are you running over HTTPS and/or localhost?",
                                         "_SUCCESS_MESSAGE_": "Image copied to clipboard!",
                                         "_IMAGE_B64_": b64encode(img.data).decode('ascii')})
        return f"""<button onclick="{onclick}" style="{COPY_BUTTON_STYLE}">{label}</button>"""

    def get_text_copy_link(self, label, text):
        """Generates HTML/JS for the copy link button"""
        onclick = read_resource_text("gofigr.resources", "copy_text.js").replace("\n", "")
        onclick = onclick.replace("_TEXT_", text)
        onclick = onclick.replace("_ALERT_ID_", self.alert_id)
        onclick = onclick.replace("_SUCCESS_MESSAGE_", "Link copied to clipboard!")
        onclick = onclick.replace("_ERROR_MESSAGE_", "Unable to copy link. Make sure your Jupyter instance is "
                                                     "running over HTTPS and/or on localhost.")
        onclick = onclick.replace("\"", "'")

        return f"""<span onclick="{onclick}" style="cursor: pointer; color: #0d53b1; ">{label}</span>"""


class DetailedWidget(RevisionWidgetBase):
    """Generates HTML/Javascript for the GoFigr Jupyter widget shown under each figure"""
    def show(self):
        """Renders this widget in Jupyter by generating the HTML/JS & calling display()"""
        logo_b64 = self.get_logo_b64()
        copy_url = self.get_text_copy_link(f"<span style='margin-left: 0.5rem;'>"
                                           f"{self.revision.figure.name}</span>"
                                           f"<span style='margin-left: 0.25rem; margin-right: 0.25rem';>"
                                           f"{FA_COPY_LIGHT}</span>",
                                           self.revision.revision_url)

        logo_html = self.get_text_copy_link(f"""<img src="data:image;base64,{logo_b64}" alt="GoFigr.io logo" 
                style='width: 3rem; height: 3rem'/>""", self.revision.revision_url)

        return display(HTML(f"""
                <div style="{WIDGET_STYLE}">
                    <div style="{ROW_STYLE + "margin-bottom: 0.5rem"}">
                    <span>Successfully published {copy_url} 
                    on {timestamp_to_local_tz(self.revision.created_on).strftime("%x at %X")}.</span> 
                    <span style="margin-left: 0.25rem;"> 
                    Revision size: {humanize.naturalsize(self.revision.size_bytes)}</span>
                    </div>

                    <div style="{ROW_STYLE}">
                    <!-- Logo -->
                    {logo_html}

                    <!-- View on GoFigr -->
                    <a href='{self.revision.revision_url}' target="_blank" style="{VIEW_BUTTON_STYLE}">{FA_VIEW}
                    <span> View on GoFigr</span></a>

                    <!-- Download -->
                    {self.get_download_link(FA_DOWNLOAD + "<span> Download</span>", True, "png")}
                    {self.get_download_link(FA_DOWNLOAD + "<span> Download (no watermark)</span>", False, "png")}

                    <!-- Copy -->
                    {self.get_copy_link(FA_COPY + "<span> Copy</span>", True, "png")}
                    {self.get_copy_link(FA_COPY + "<span> Copy (no watermark)</span>", False, "png")}

                    </div>

                    <div id={self.alert_id} style="{ROW_STYLE + MESSAGE_STYLE}">
                    </div>
                </div>"""))


class CompactWidget(RevisionWidgetBase):
    """Generates a compact GoFigr widget"""

    def show(self):
        """Renders this widget in Jupyter by generating the HTML/JS & calling display()"""
        logo_b64 = self.get_logo_b64()
        logo_html = self.get_text_copy_link(f"""<img src="data:image;base64,{logo_b64}" alt="GoFigr.io logo" 
        style='width: 2rem; height: 2rem'/>""", self.revision.revision_url)

        return display(HTML(f"""
                <div style="{COMPACT_WIDGET_STYLE}">
                    <div style="{ROW_STYLE + "margin-bottom: 0.0rem"}">
                    <!-- Logo -->
                    {logo_html}               

                    <!-- View on GoFigr -->
                    <a href='{self.revision.revision_url}' target="_blank" style="{VIEW_BUTTON_STYLE}">{FA_VIEW}
                    <span> View on GoFigr</span></a>

                    <!-- Download -->
                    {self.get_download_link(FA_DOWNLOAD + "<span> Download</span>", True, "png")}
                    {self.get_download_link(FA_DOWNLOAD + "<span> Download (no watermark)</span>", False, "png")}

                    <!-- Copy -->
                    {self.get_copy_link(FA_COPY + "<span> Copy</span>", True, "png")}
                    {self.get_copy_link(FA_COPY + "<span> Copy (no watermark)</span>", False, "png")}

                    </div>

                    <div id={self.alert_id} style="{ROW_STYLE + MESSAGE_STYLE}">
                    </div>
                </div>"""))


class MinimalWidget(RevisionWidgetBase):
    """Generates a compact GoFigr widget"""

    def show(self):
        """Renders this widget in Jupyter by generating the HTML/JS & calling display()"""
        logo_b64 = self.get_logo_b64()
        logo_html = self.get_text_copy_link(f"""<img src="data:image;base64,{logo_b64}" alt="GoFigr.io logo" 
        style='width: 2rem; height: 2rem; margin-right: 0.5rem;'/>""", self.revision.revision_url)

        return display(HTML(f"""
                <div style="{COMPACT_WIDGET_STYLE}">
                    <div style="{ROW_STYLE + "margin-bottom: 0.0rem"}">
                    <!-- Logo -->
                    {logo_html}               

                    <!-- View on GoFigr -->
                    <a href='{self.revision.revision_url}' target="_blank" style="margin-top: auto; margin-bottom: auto;">View on GoFigr</a>
                    </div>
                    
                    <div id={self.alert_id} style="{ROW_STYLE + MESSAGE_STYLE}">
                    </div>
                </div>"""))


class StartupWidget(WidgetBase):
    """Generates a GoFigr startup widget"""
    def __init__(self, extension):
        super().__init__()
        self.extension = extension

    def get_link(self, obj):
        """Gets the app link to a GoFigr object"""
        return f"""<a style="margin-left: 0.25em" href={obj.app_url}>{obj.name}</a>"""

    def show(self):
        """Renders this widget in Jupyter by generating the HTML/JS & calling display()"""
        self.extension.startup_widget_shown = True

        logo_b64 = self.get_logo_b64()
        logo_html = f"""<img src="data:image;base64,{logo_b64}" alt="GoFigr.io logo" 
            style='width: 2rem; height: 2rem; margin-right: 0.5rem;'/>"""

        return display(HTML(f"""
                    <div style="{WIDGET_STYLE}">
                        <div style="{ROW_STYLE + "margin-bottom: 0.0rem"}">
                            <!-- Logo -->
                            {logo_html}      
                            
                            <div style="margin-top: auto; margin-bottom: auto;">
                                <span style="font-weight: bold">GoFigr active</span>. 
                                Workspace: {self.get_link(self.extension.publisher.workspace)}. 
                                Analysis: {self.get_link(self.extension.publisher.analysis)}. 
                                Autopublish: {self.extension.auto_publish}.
                            </div>
                        </div>
                        
                    </div>"""))


class LiteStartupWidget(WidgetBase):
    """Generates a GoFigr startup widget"""

    def __init__(self, extension):
        super().__init__()
        self.extension = extension

    def show(self):
        """Renders this widget in Jupyter by generating the HTML/JS & calling display()"""
        logo_b64 = self.get_logo_b64()
        logo_html = f"""<img src="data:image;base64,{logo_b64}" alt="GoFigr.io logo" 
            style='width: 2rem; height: 2rem; margin-right: 0.5rem;'/>"""

        meta = NotebookMetadataAnnotator().parse_metadata(error=False)
        if meta is None:
            notebook_name = "N/A"
        else:
            notebook_name = meta.get(NOTEBOOK_NAME, "N/A")

        return display(HTML(f"""
                    <div style="{WIDGET_STYLE}">
                        <div style="{ROW_STYLE + "margin-bottom: 0.0rem"}">
                            <!-- Logo -->
                            {logo_html}      

                            <div style="margin-top: auto; margin-bottom: auto;">
                                <span style="font-weight: bold">GoFigr Lite active. Notebook: </span>
                                <code>{notebook_name}</code>. 
                            </div>                            
                        </div>                        
                        
                        <div style="{ROW_STYLE + "margin-bottom: 0.0rem; margin-top: 1.0rem"}">
                            <span>GoFigr Lite adds annotation to your figures but does not publish them to our servers. To experience 
                            full functionality, consider using the full GoFigr extension. See 
                            <a href="https://gofigr.io/gofigr-lite/">this page</a> for details.
                            </span>                
                        </div>

                    </div>"""))


class AssetWidget(WidgetBase):
    """Generates a GoFigr startup widget"""
    def __init__(self, rev):
        super().__init__()
        self.rev = rev

    def get_link(self, obj):
        """Gets the app link to a GoFigr object"""
        if not obj.asset.name:
            obj.asset.fetch()
        return f"""<a style="margin-left: 0.25em" href={obj.app_url}>{obj.asset.name}</a>"""

    def show(self):
        """Renders this widget in Jupyter by generating the HTML/JS & calling display()"""
        logo_b64 = self.get_logo_b64()
        logo_html = f"""<img src="data:image;base64,{logo_b64}" alt="GoFigr.io logo" 
            style='width: 2rem; height: 2rem; margin-right: 0.5rem;'/>"""

        return display(HTML(f"""
                    <div style="{WIDGET_STYLE}">
                        <div style="{ROW_STYLE + "margin-bottom: 0.0rem"}">
                            <!-- Logo -->
                            {logo_html}      

                            <div style="margin-top: auto; margin-bottom: auto;">
                                <span style="font-weight: bold">Asset: </span>
                                {self.get_link(self.rev)} {"(new revision)" if self.rev.is_new_revision else "(existing revision)"}.
                            </div>
                        </div>

                    </div>"""))



"""\
Copyright (c) 2022, Flagstaff Solutions, LLC
All rights reserved.

"""
# pylint: disable=cyclic-import, no-member, global-statement, protected-access
# pylint: disable=too-many-locals, ungrouped-imports

import io
import logging
import os
import pickle
import sys
import traceback
from functools import wraps
from pathlib import Path
from uuid import UUID

from IPython import get_ipython

import gofigr
from gofigr import GoFigr, API_URL
from gofigr.annotators import NotebookMetadataAnnotator, NOTEBOOK_NAME, NOTEBOOK_PATH, IPythonAnnotator
from gofigr.publisher import Publisher, DEFAULT_ANNOTATORS, DEFAULT_BACKENDS, _mark_as_published, is_published, \
    is_suppressed
from gofigr.proxy import run_proxy_async, get_javascript_loader
from gofigr.profile import MeasureExecution
from gofigr.trap import GfDisplayPublisher, SuppressDisplayTrap
from gofigr.utils import from_config_or_env
from gofigr.widget import DetailedWidget, StartupWidget

try:
    from IPython.core.display_functions import display
except ModuleNotFoundError:
    from IPython.core.display import display


logger = logging.getLogger(__name__)


EXTENSION_NAME = "_GF_EXTENSION"

# pylint: disable=too-many-instance-attributes
class _GoFigrExtension:
    """\
    Implements the main Jupyter extension functionality. You will not want to instantiate this class directly.
    Instead, please call get_extension().
    """
    def __init__(self, ip,
                 auto_publish=False,
                 notebook_metadata=None,
                 configured=False,
                 loader_shown=False,
                 asset_log=None):
        """\

        :param ip: iPython shell instance
        :param auto_publish: whether to auto-publish figures
        :param pre_run_hook: function to use as a pre-run hook
        :param post_execute_hook: function to use as a post-execute hook
        :param notebook_metadata: information about the running notebook, as a key-value dictionary
        :param offline: if True, will provide watermarking but will not publish to GoFigr.io.

        """
        self.shell = ip
        self.auto_publish = auto_publish
        self.cell = None
        self.proxy = None
        self.loader_shown = loader_shown
        self.configured = configured
        if notebook_metadata is None:
            self.notebook_metadata = NotebookMetadataAnnotator().try_get_metadata()
        else:
            self.notebook_metadata = notebook_metadata

        self.gf = None  # active GF object
        self.publisher = None  # current Publisher instance
        self.wait_for_metadata = None  # callable which waits for metadata to become available
        self.asset_log = asset_log if asset_log is not None else {}
        self.startup_widget_shown = False

        self.deferred_revisions = []

    @property
    def is_ready(self):
        """True if the extension has been configured and ready for use."""
        return self.configured and self.notebook_metadata is not None

    def resolve_analysis(self):
        """Gets the current analysis"""
        if isinstance(self.publisher.analysis, gofigr.NotebookName):
            meta = NotebookMetadataAnnotator().try_get_metadata()
            if meta is not None:
                self.publisher.analysis = self.publisher.workspace.get_analysis(name=Path(meta[NOTEBOOK_NAME]).stem,
                                                                                create=True)
                self.publisher.analysis.fetch()

        return self.publisher.analysis

    def display_trap(self, data, suppress_display):
        """\
         Called whenever *any* code inside the Jupyter session calls display().
        :param data: dictionary of MIME types
        :param suppress_display: callable with no arguments. Call to prevent the originating figure from being shown.
        :return: None

        """
        if self.publisher is None:
            logger.warning("No publisher configured.")
            return

        if self.auto_publish:
            self.publisher.auto_publish_hook(self, data, suppress_display)

    def add_to_deferred(self, rev):
        """\
        Adds a revision to a list of deferred revisions. Such revisions will be annotated in the post_run_cell
        hook, and re-saved.

        This functionality exists because it's possible to load the GoFigr extension and publish figures in the same
        cell, in which case GoFigr will not receive the pre_run_cell hook and will not have access to cell information
        when the figure is published. This functionality allows us to obtain the cell information after it's run
        (in the post_run_cell hook), re-run annotators, and update the figure with full annotations.

        :param rev: revision to defer
        :return: None
        """
        if rev not in self.deferred_revisions:
            self.deferred_revisions.append(rev)

    def check_config(self):
        """Ensures the plugin has been configured for use"""
        if not self.configured:
            raise RuntimeError("GoFigr not configured. Please call configure() first.")

    def pre_run_cell(self, info):
        """\
        Default pre-run cell hook.

        :param info: Cell object
        :return:None

        """
        self.cell = info

    def _get_metadata_from_proxy(self, result):
        if self.configured and not self.loader_shown and "_VSCODE" not in result.info.raw_cell:
            self.proxy, self.wait_for_metadata = run_proxy_async(self.gf, proxy_callback)

            with SuppressDisplayTrap():
                display(get_javascript_loader(self.gf, self.proxy))
                self.loader_shown = True

        if self.notebook_metadata is None and self.wait_for_metadata is not None:
            self.wait_for_metadata()
            self.wait_for_metadata = None


    def post_run_cell(self, result):
        """Post run cell hook.

        :param result: ExecutionResult
        :return: None

        """
        self.cell = result.info

        if self.notebook_metadata is None:
            self._get_metadata_from_proxy(result)

        self.resolve_analysis()
        if self.is_ready and not self.startup_widget_shown:
            StartupWidget(get_extension()).show()

        while len(self.deferred_revisions) > 0:
            rev = self.deferred_revisions.pop(0)
            rev = self.publisher.annotate(rev)
            rev.save(silent=True)

        self.cell = None

    def _register_handler(self, event_name, handler):
        """Inserts a handler at the beginning of the list while avoiding double-insertions"""
        handlers = [handler]
        for hnd in self.shell.events.callbacks[event_name]:
            self.shell.events.unregister(event_name, hnd)
            if hnd != handler:  # in case it's already registered, skip it
                handlers.append(hnd)

        for hnd in handlers:
            self.shell.events.register(event_name, hnd)

    def unregister(self):
        """\
        Unregisters all hooks, effectively disabling the plugin.

        """
        try:
            self.shell.events.unregister('pre_run_cell', self.pre_run_cell)
        except ValueError:
            pass

        try:
            self.shell.events.unregister('post_run_cell', self.post_run_cell)
        except ValueError:
            pass

    def register_hooks(self):
        """\
        Register all hooks with Jupyter.

        :return: None
        """
        self._register_handler('pre_run_cell', self.pre_run_cell)
        self._register_handler('post_run_cell', self.post_run_cell)

        native_display_publisher = self.shell.display_pub
        if not isinstance(native_display_publisher, GfDisplayPublisher):
            self.shell.display_pub = GfDisplayPublisher(native_display_publisher, display_trap=self.display_trap)


def require_configured(func):
    """\
    Decorator which throws an exception if configure() has not been called yet.

    :param func:
    :return:
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        ext = _get_extension_nocheck()
        if ext is None:
            raise RuntimeError("Please load the extension: %load_ext gofigr")

        ext.check_config()
        return func(*args, **kwargs)

    return wrapper


@require_configured
def get_extension():
    """Returns the GoFigr Jupyter extension instance"""
    return get_ipython().user_ns.get(EXTENSION_NAME)


def _get_extension_nocheck():
    """Returns the GoFigr Jupyter extension instance"""
    return get_ipython().user_ns.get(EXTENSION_NAME)



def _load_ipython_extension(ip):
    """\
    Loads the Jupyter extension. Aliased to "load_ipython_extension" (no leading underscore) in the main init.py file.

    :param ip: IPython shell
    :return: None

    """
    ext = _get_extension_nocheck()
    if ext is not None:
        ext.unregister()

    ext = _GoFigrExtension(ip)
    ip.user_ns[EXTENSION_NAME] = ext
    ext.register_hooks()

    try:
        configure()
    except Exception as e:  # pylint: disable=broad-exception-caught
        if "auth" in str(e).lower() and "failed" in str(e).lower():
            print("GoFigr authentication failed. Please manually call configure(api_key=<YOUR API KEY>).",
                  file=sys.stderr)
        else:
            print(traceback.format_exc(), file=sys.stderr)
            print(f"Could not automatically configure GoFigr. Please call configure() manually. Error: {e}",
                  file=sys.stderr)

    ip.user_ns["configure"] = configure
    ip.user_ns["get_extension"] = get_extension
    ip.user_ns["publish"] = publish

    # These have to use the fully qualified reference or isinstance won't work correctly
    # after the extension is re-loaded. That's because gofigr.FindByName and gofigr.jupyter.FindByName (imported scope)
    # will contain different instantiations of the type.
    ip.user_ns["FindByName"] = gofigr.FindByName
    ip.user_ns["ApiId"] = gofigr.ApiId
    ip.user_ns["NotebookName"] = gofigr.NotebookName


def parse_uuid(val):
    """\
    Attempts to parse a UUID, returning None if input is not a valid UUID.

    :param val: value to parse
    :return: UUID (as a string) or None

    """
    try:
        return str(UUID(val))
    except ValueError:
        return None


def proxy_callback(result):
    """Proxy callback"""
    if result is not None and hasattr(result, 'metadata'):
        get_extension().notebook_metadata = result.metadata


class JupyterPublisher(Publisher):
    """\
    Adds Jupyter-specific functionality to GoFigr's base Publisher class.

    """
    def __init__(self, *args, clear=True, **kwargs):
        super().__init__(*args, clear=clear, **kwargs)

    @property
    def _non_ipython_annotators(self):
        return [ann for ann in self.annotators if not isinstance(ann, IPythonAnnotator)]

    def auto_publish_hook(self, extension, data, suppress_display=None):
        """\
        Hook for automatically publishing figures without an explicit call to publish().

        :param extension: GoFigrExtension instance
        :param data: data being published. This will usually be a dictionary of mime formats.
        :param suppress_display: if used in an auto-publish hook, this will contain a callable which will
               suppress the display of this figure using the native IPython backend.

        :return: None
        """
        for backend in self.backends:
            compatible_figures = list(backend.find_figures(extension.shell, data))
            for fig in compatible_figures:
                if not is_published(fig) and not is_suppressed(fig):
                    self.publish(fig=fig, backend=backend, suppress_display=suppress_display)

            if len(compatible_figures) > 0:
                break

    def publish(self, *args, **kwargs):
        with SuppressDisplayTrap():
            ext = get_extension()

            if ext.cell is None:
                revision = super().publish(*args, annotators=self._non_ipython_annotators, **kwargs)
                ext.add_to_deferred(revision)
            else:
                revision = super().publish(*args, **kwargs)
            return revision


# pylint: disable=too-many-arguments, too-many-locals
@from_config_or_env("GF_", gofigr.find_config())
def configure(username=None,
              password=None,
              api_key=None,
              workspace=None,
              analysis=gofigr.NotebookName(),
              url=API_URL,
              default_metadata=None, auto_publish=True,
              watermark=None, annotators=DEFAULT_ANNOTATORS,
              notebook_name=None, notebook_path=None,
              backends=DEFAULT_BACKENDS,
              widget_class=DetailedWidget,
              save_pickle=True,
              show_watermark=True):
    """\
    Configures the Jupyter plugin for use.

    :param username: GoFigr username (if used instead of API key)
    :param password: GoFigr password (if used instead of API key)
    :param api_key: API Key (if used instead of username and password)
    :param url: API URL
    :param workspace: one of: API ID (string), ApiId instance, or FindByName instance
    :param analysis: one of: API ID (string), ApiId instance, FindByName, or NotebookName instance
    :param default_metadata: dictionary of default metadata values to save for each revision
    :param auto_publish: if True, all figures will be published automatically without needing to call publish()
    :param watermark: custom watermark instance (e.g. DefaultWatermark with custom arguments)
    :param annotators: list of annotators to use. Default: DEFAULT_ANNOTATORS
    :param notebook_name: name of the notebook (if you don't want it to be inferred automatically)
    :param notebook_path: path to the notebook (if you don't want it to be inferred automatically)
    :param backends: backends to use (e.g. MatplotlibBackend, PlotlyBackend)
    :param widget_class: Widget type to show, e.g. DetailedWidget or CompactWidget. It will appear below the
        published figure
    :param save_pickle: if True, will save the figure in pickle format in addition to any of the image formats
    :param show_watermark: True to show watermarked figures instead of original.
        False to always display the unmodified figure. Default True.
    :return: None

    """
    extension = _get_extension_nocheck()
    if extension is None:
        raise RuntimeError("Please load the extension: %load_ext gofigr")

    if isinstance(auto_publish, str):
        auto_publish = auto_publish.lower() == "true"  # in case it's coming from an environment variable

    with MeasureExecution("Login"):
        gf = GoFigr(username=username, password=password, url=url, api_key=api_key,
                    asset_log=extension.asset_log)

    if default_metadata is None:
        default_metadata = {}

    if notebook_path is not None:
        default_metadata[NOTEBOOK_PATH] = notebook_path

    if notebook_name is not None:
        default_metadata[NOTEBOOK_NAME] = notebook_name

    if str(analysis) == "NotebookName":  # str in case it's from config/env
        analysis = gofigr.NotebookName()

    worx = gf.find_workspace(workspace)
    gf.workspace_id = worx.api_id
    publisher = JupyterPublisher(gf,
                                 workspace=worx,
                                 analysis=gf.find_analysis(worx, analysis),
                                 default_metadata=default_metadata,
                                 watermark=watermark,
                                 annotators=[make_annotator() for make_annotator in annotators],
                                 backends=backends,
                                 widget_class=widget_class,
                                 save_pickle=save_pickle,
                                 show_watermark=show_watermark)
    extension.gf = gf
    extension.publisher = publisher
    extension.auto_publish = auto_publish
    extension.loader_shown = False
    extension.configured = True
    extension.resolve_analysis()

    extension.shell.user_ns["gf"] = gf

    if extension.is_ready:
        StartupWidget(extension).show()


def _base_publish(ext, fig=None, backend=None, **kwargs):
    """
    Publishes a figure using the provided extension and backend. If neither a figure
    nor a backend is supplied, it publishes default figures across all available
    backends supported by the extension. The function handles publication based on
    the provided or determined inputs.

    :param ext: The extension object containing the publisher to handle the publishing process.
    :param fig: Optional; The figure object to be published. If None, the method attempts
        to use default figures from the available backends.
    :param backend: Optional; The backend to be used for publishing. If None, the method
        attempts to publish using all backends available within the extension's publisher.
    :param kwargs: Additional keyword arguments to customize the publishing process.
    :return: None if no figure is provided or publish process fails; otherwise returns
        the result of the publish process as managed by the extension's publisher.

    """
    if fig is None and backend is None:
        # If no figure and no backend supplied, publish default figures across all available backends
        for available_backend in ext.publisher.backends:
            fig = available_backend.get_default_figure(silent=True)
            if fig is not None:
                return ext.publisher.publish(fig=fig, backend=available_backend, **kwargs)

        return None
    else:
        return ext.publisher.publish(fig=fig, backend=backend, **kwargs)


@require_configured
def publish(fig=None, backend=None, **kwargs):
    """\
    Publishes a figure. See :func:`gofigr.jupyter.Publisher.publish` for a list of arguments. If figure and backend
    are both None, will publish default figures across all available backends.

    :param fig: figure to publish
    :param backend: backend to use
    :param kwargs:
    :return:

    """
    return _base_publish(ext=get_extension(), fig=fig, backend=backend, **kwargs)



@require_configured
def get_gofigr():
    """Gets the active GoFigr object."""
    return get_extension().gf


@require_configured
def load_pickled_figure(api_id):
    """\
    Unpickles a GoFigr revision and returns it as a backend-specific Python object, e.g. a plt.Figure if
    the figure was generated with matplotlib. Throws a RuntimeException if the figure is not found or does
    not have pickle data.

    :param api_id: API ID of the revision
    :return: backend-dependent figure object, e.g. plt.Figure().

    """
    gf = get_gofigr()
    rev = gf.Revision(api_id=api_id).fetch(fetch_data=False)
    for data in rev.image_data:
        if data.format == "pickle":
            data.fetch()

            fig = pickle.load(io.BytesIO(data.data))
            return _mark_as_published(fig)

    raise RuntimeError("This revision doesn't have pickle data.")


@require_configured
def download_file(api_id, file_name, path):
    """\
    Downloads a file and saves it.

    :param api_id: API ID of the revision containing the file
    :param file_name: name of the file to download
    :param path: where to save the file (either existing directory or full path with the file name)
    :return: number of bytes written

    """
    if os.path.exists(path) and os.path.isdir(path):
        path = os.path.join(path, file_name)

    gf = get_gofigr()
    rev = gf.Revision(api_id=api_id).fetch(fetch_data=False)
    for data in rev.file_data:
        if data.name == file_name:
            data.fetch()
            data.write(path)
            return len(data.data)

    raise RuntimeError(f"Could not find file \"{file_name}\" in revision {api_id}.")


@require_configured
def download_all(api_id, path):
    """\
    Downloads all files attached to a revision

    :param api_id: API ID of the revision containing the files
    :param path: directory where to save the files
    :return: number of bytes written

    """
    if not os.path.exists(path) or not os.path.isdir(path):
        raise RuntimeError(f"{path} does not exist or is not a directory")

    gf = get_gofigr()
    rev = gf.Revision(api_id=api_id).fetch(fetch_data=False)
    num_bytes = 0
    for data in rev.file_data:
        data.fetch()
        data.write(os.path.join(path, data.name))
        num_bytes += len(data.data)

    return num_bytes



"""\
Copyright (c) 2024, Flagstaff Solutions, LLC
All rights reserved.

"""
import inspect
import io
import sys

import plotnine

from gofigr.backends import GoFigrBackend, get_all_function_arguments
from gofigr.backends.matplotlib import SuppressPltWarnings


class PlotnineBackend(GoFigrBackend):
    """\
    MatplotLib backend for GoFigr.

    """
    def get_backend_name(self):
        return "plotnine"

    def is_compatible(self, fig):
        return isinstance(fig, plotnine.ggplot)

    def is_interactive(self, fig):
        return False

    def find_figures(self, shell, data):
        frames = inspect.stack()
        # Walk through the stack in *reverse* order (from top to bottom), to find the first call
        # in case display() was called recursively
        for f in reversed(frames):
            if ("ipython_display" in f.function or "draw" in f.function) and f.filename.endswith("ggplot.py"):
                for arg_value in get_all_function_arguments(f):
                    if self.is_compatible(arg_value):
                        yield arg_value

                break

    def get_default_figure(self, silent=False):
        if not silent:
            print("plotnine does not have a default figure. Please specify a figure to publish.", file=sys.stderr)

    def get_title(self, fig):
        return fig.labels.title

    def figure_to_bytes(self, fig, fmt, params):
        with SuppressPltWarnings():
            bio = io.BytesIO()
            fig.save(bio, format=fmt, verbose=False, **params)

            bio.seek(0)
            return bio.read()

    def close(self, fig):
        pass

    def get_supported_image_formats(self):
        return ["png", "eps", "pdf", "jpeg", "jpg", "svg", "tiff"]



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
import os
import sys
sys.path.insert(0, os.path.abspath('..'))

project = 'GoFigr Client Library'
copyright = '2024, Flagstaff Solutions, LLC'
author = 'Maciej Pacula'
release = '1.3.1'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.viewcode',
    'sphinx.ext.napoleon',
    "sphinxcontrib.jquery"
]

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', 'tests', 'venv', '.*egg.*', "dist"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = []



"""\
Copyright (c) 2023, Flagstaff Solutions, LLC
All rights reserved.

"""
import sys
from argparse import ArgumentParser

import numpy as np
import pandas as pd
import json
import os
import re


def find_pkg(name, lines):
    """Given output of pip freeze, finds the exact package version"""
    m = [x for x in lines if x.startswith(f"{name}==")]
    if len(m) == 0:
        return None
    elif len(m) == 1:
        return re.match(r'.*==(.*)', m[0]).group(1)
    else:
        raise ValueError(f'Ambiguous: {name}. Matches: {m}')


def parse_results(path, results_name):
    """Parses results of a single integration test and returns them as a data frame"""
    if os.path.exists(os.path.join(path, results_name)):
        with open(os.path.join(path, results_name), 'r') as f:
            results = json.load(f)
    else:
        results = {'platform': 'N/A',
                   'results': [{'name': 'N/A'}]}

    platform = results['platform']
    tests = results['results']

    df = pd.DataFrame(tests)
    df['platform'] = platform

    if os.path.exists(os.path.join(path, 'pip_freeze.txt')):
        with open(os.path.join(path, 'pip_freeze.txt'), 'r') as f:
            packages = [x.strip() for x in f.readlines()]
    else:
        packages = []

    df['notebook'] = find_pkg("notebook", packages)
    df['jupyterlab'] = find_pkg("jupyterlab", packages)
    df['jupyter_server'] = find_pkg("jupyter_server", packages)
    df['jupyter_core'] = find_pkg("jupyter_core", packages)
    df['jupyter_client'] = find_pkg("jupyter_client", packages)
    df['ipython'] = find_pkg("ipython", packages)
    df['ipykernel'] = find_pkg("ipykernel", packages)
    df['matplotlib'] = find_pkg("matplotlib", packages)
    df['plotly'] = find_pkg("plotly", packages)
    df['plotnine'] = find_pkg("plotnine", packages)

    with open(os.path.join(path, "config.json"), 'r') as f:
        config = json.load(f)

    if os.path.exists(os.path.join(path, "python_version.txt")):
        with open(os.path.join(path, "python_version.txt"), 'r') as f:
            txt = f.read()
            df['python'] = re.match(r'Python\s+([\d\.]+)', txt).group(1)
            df['python_minor_version'] = int(re.match(r'Python\s+3\.(\d+)\..*', txt).group(1))
    else:
        df['python'] = 'N/A'
        df['python_minor_version'] = 'N/A'

    df['service'] = config['service']
    df['name'] = config['name']

    return df


def one(xs):
    """Makes sure there's exactly one unique value in a list and returns it"""
    xs = list(set(xs))
    if len(xs) == 0:
        return None
    elif len(xs) == 1:
        return xs[0]
    else:
        raise ValueError(xs)


_TEST_COLUMNS = ['number_of_revisions',
                'notebook_name',
                'notebook_path',
                'image_png',
                'image_png_watermark',
                'image_eps',
                'image_svg',
                'image_html',
                'file_pickle',
                'num_assets',
                'asset_figure',
                'asset_rev_not_null',
                'asset_name',
                'text',
                'cell_code',
                'backend',
                'history']

_LITE_TEST_COLUMNS = ['notebook_name',
                     'notebook_path',
                     'extension_version',
                     'metadata',
                     'comm',
                     'title',
                     'image_not_null',
                     'image_type_check']

_COLUMN_ORDER = ['name',
                'platform',
                'service',
                'error',
                'python',
                'python_minor_version',

                'notebook',
                'jupyterlab',
                'jupyter_server',
                'jupyter_core',
                'jupyter_client',
                'ipython',
                'ipykernel',
                'matplotlib',
                'plotly',
                'test_name',
                'error',
                'elapsed_seconds']


def get_columns(test):
    if test == "full":
        return _TEST_COLUMNS
    elif test == "lite":
        return _LITE_TEST_COLUMNS
    else:
        raise ValueError(test)


def get_column_order(test):
    if test == "full":
        return _COLUMN_ORDER + get_columns(test)
    elif test == "lite":
        return _COLUMN_ORDER + get_columns(test)
    else:
        raise ValueError(test)


def abbreviate(text, max_len=100):
    if len(text) <= max_len:
        return text
    else:
        return text[0:max_len - 3] + "..."


def parse_version(text):
    """Parses a major.minor.patch version string"""
    return tuple([int(x) for x in text.split(".")])


def summarize_results(df, test):
    """Summarizes test results"""
    all_tests = []
    passed_tests = []
    failed_tests = []

    if 'test_name' not in df.columns:  # none of the tests have run
        failed_tests.append("None of the tests have run")
    else:
        for _, row in df.iterrows():
            test_name = row['test_name']
            is_plotly = "plotly" in test_name.lower()
            is_py3dmol = 'py3dmol' in test_name.lower()
            is_matplotlib = 'mpl' in test_name.lower()
            is_plotnine = "plotnine" in test_name.lower()

            for col in get_columns(test):
                check_name = f"{test_name}>{col}"
                if row.get(col) is True:  # test passed
                    passed_tests.append(check_name)
                    all_tests.append(check_name)
                elif (is_matplotlib or is_plotnine) and col in ['image_html', 'image_html_watermark']:
                    pass  # not an interactive backend
                elif is_matplotlib and "Anonymous" in test_name and col == "number_of_revisions":
                    # Because figures aren't cleared between re-attempts of the same config, Anonymous fig can end up
                    # with more than one revision. This needs to be fixed, but here's a temporary workaround.
                    pass
                elif is_plotnine and col in ['file_pickle'] and parse_version(row['plotnine']) < (0, 14, 0):
                    pass
                elif is_plotly and col in ["image_eps"]:  # plotly doesn't support EPS, so this failure is expected
                    pass
                elif is_py3dmol and (col in ["image_svg", "image_eps"] or "3.7" in row["python"]): # similar for py3dmol
                    pass
                else:
                    failed_tests.append(check_name)
                    all_tests.append(check_name)

    return pd.DataFrame({'name': [one(df['name'])],
                         'service': [one(df['service'])],
                         'python': [one(df['python'])],
                         'elapsed_seconds': [np.round(df['elapsed_seconds'].max(), 1)
                                             if 'elapsed_seconds' in df.columns else "N/A"],
                         'result': ["success" if len(failed_tests) == 0 else "failed"],
                         'all_tests': [len(all_tests)],
                         'passed_tests': [len(passed_tests)],
                         'failed_tests': [len(failed_tests)],
                         'failed_tests_detail': [abbreviate(", ".join(failed_tests))]})


def summarize_all(path, single, results_name):
    """Finds all integration tests in a directory, summarizes them, and returns the combined dataframe"""
    if 'lite' in results_name:
        test = "lite"
    else:
        test = "full"

    frames = []
    summary_frames = []
    if single:
        print(f"{os.path.basename(path)}...")
        df = parse_results(path, results_name)
        frames.append(df)
        summary_frames.append(summarize_results(df, test))
    else:
        for name in os.listdir(path):
            if test == 'lite' and 'lite' not in name:
                continue

            if test == "full" and 'lite' in name:
                continue

            full = os.path.join(path, name)
            if os.path.isdir(full):
                print(f"{name}...")
                df = parse_results(full, results_name)
                frames.append(df)
                summary_frames.append(summarize_results(df, test))

    df = pd.concat(frames, ignore_index=True).sort_values(by=['python_minor_version', 'service'])
    column_subset = [x for x in get_column_order(test) if x in df.columns]
    test_column_subset = [x for x in get_columns(test) if x in df.columns]
    df.loc[:, test_column_subset] = df[test_column_subset].fillna(value='N/A')

    df_summary = pd.concat(summary_frames, ignore_index=True)
    return df[column_subset], df_summary


def main():
    """Main"""
    parser = ArgumentParser(description="Generates a compatibility report")
    parser.add_argument("directory", help="Directory containing integration test results")
    parser.add_argument("output", help="Where to save the output Excel file")
    parser.add_argument("detailed_output", help="Where to save the detailed Excel file")
    parser.add_argument("--single", help="Processes a single run only", action="store_true")
    parser.add_argument("--results", help="Name of the results file", default="integration_test.json")
    args = parser.parse_args()

    df_detail, df_summary = summarize_all(args.directory, single=args.single, results_name=args.results)
    print("\n=== Result summary ===")
    print(df_summary)
    print("\n")

    df_summary.to_excel(args.output, index=False)
    print(f"Saved to {args.output}")

    df_detail.to_excel(args.detailed_output, index=False)
    print(f"Saved detailed output to {args.detailed_output}")

    df_failed = df_summary[df_summary['result'] != "success"]
    df_success = df_summary[df_summary['result'] == "success"]
    print(f"\n\n  => Ran {len(df_summary)} tests. {len(df_success)} successes. {len(df_failed)} failures.")

    if (df_summary['result'] != "success").any():
        print(f"\n  => Tests failed.", file=sys.stderr)
        sys.exit(1)
    else:
        print("  => All tests successful.")


if __name__ == "__main__":
    main()



"""\
Copyright (c) 2022-2025, Flagstaff Solutions, LLC
All rights reserved.

"""

class UnauthorizedError(RuntimeError):
    """\
    Thrown if user doesn't have permissions to perform an action.
    """
    pass


class MethodNotAllowedError(RuntimeError):
    """\
    Thrown if a given REST action is not supported/allowed.
    """
    pass



"""\
Copyright (c) 2022, Flagstaff Solutions, LLC
All rights reserved.

"""
# pylint: disable=no-member

import contextlib
import inspect
import logging
import warnings
from json import dumps as json_dumps

import requests
from PIL import Image
from requests import Session

from gofigr.exceptions import UnauthorizedError, MethodNotAllowedError
from gofigr.models import *
from gofigr.utils import from_config_or_env, try_parse_uuid4
from gofigr.widget import AssetWidget

LOGGER = logging.getLogger(__name__)

API_URL = "https://api.gofigr.io"
API_VERSION = "v1.2"

APP_URL = "https://app.gofigr.io"

PANDAS_READERS = ["read_csv", "read_excel", "read_json", "read_html", "read_parquet", "read_feather",
                  "read_hdf", "read_pickle", "read_sas"]
REVISION_ATTR = "_gofigr_revision"


def _jupyter_labextension_paths():
    return [{
        "src": "labextension",
        "dest": "gofigr"
    }]


def assert_one(elements, error_none=None, error_many=None):
    """\
    Asserts that a list/tuple contains only a single element (raising an exception if not), and returns
    that element.

    :param elements: list/tuple
    :param error_none: error message if input is empty
    :param error_many: error message if multiple elements are present
    :return: the single element in the input
    """
    if len(elements) == 0:
        raise ValueError(error_none or "Expected exactly one value but got none")
    elif len(elements) > 1:
        raise ValueError(error_many or f"Expected exactly one value but got n={len(elements)}")
    else:
        return elements[0]


class UserInfo:
    """\
    Stores basic information about a user: username, email, etc.

    """
    def __init__(self, username, first_name, last_name, email, date_joined, is_active, avatar,
                 is_staff, user_profile):
        """\

        :param username:
        :param first_name:
        :param last_name:
        :param email:
        :param date_joined:
        :param is_active:
        :param avatar: avatar as a PIL.Image instance
        :param is_staff: whether the user is staff or not
        """
        self.username = username
        self.first_name, self.last_name = first_name, last_name
        self.email = email
        self.date_joined = date_joined
        self.is_active = is_active
        self.avatar = avatar
        self.is_staff = is_staff
        self.user_profile = user_profile

    @staticmethod
    def _avatar_to_b64(img):
        if not img:
            return None

        bio = io.BytesIO()
        img.save(bio, format="png")
        return b64encode(bio.getvalue()).decode('ascii')

    @staticmethod
    def _avatar_from_b64(data):
        if not data:
            return None

        return Image.open(io.BytesIO(b64decode(data)))

    @staticmethod
    def from_json(obj):
        """\
        Parses a UserInfo object from JSON

        :param obj: JSON representation
        :return: UserInfo instance
        """
        date_joined = obj.get('date_joined')
        return UserInfo(username=obj.get('username'),
                        first_name=obj.get('first_name'),
                        last_name=obj.get('last_name'),
                        email=obj.get('email'),
                        date_joined=dateutil.parser.parse(date_joined) if date_joined is not None else None,
                        is_active=obj.get('is_active'),
                        is_staff=obj.get('is_staff'),
                        user_profile=obj.get('user_profile', {}),
                        avatar=UserInfo._avatar_from_b64(obj.get('avatar')))

    def to_json(self):
        """Converts this UserInfo object to json"""
        return {'username': self.username,
                'first_name': self.first_name,
                'last_name': self.last_name,
                'email': self.email,
                'date_joined': str(self.date_joined) if self.date_joined else None,
                'is_active': self.is_active,
                'is_staff': self.is_staff,
                'user_profile': self.user_profile,
                'avatar': UserInfo._avatar_to_b64(self.avatar)}

    def __str__(self):
        return json_dumps(self.to_json())

    def __eq__(self, other):
        return str(self) == str(other)


def find_config(current_dir=None, filename=".gofigr"):
    """\
    Recursively searches for the GoFigr configuration file starting in current_dir, then walking up
    the directory hierarchy. If one is not found, we then check the user's home directory.

    :param current_dir: start directory. Defaults to current directory.
    :param filename: filename to look for. Defaults to .gofigr.
    :return: path if found, or None

    """
    if current_dir is None:
        current_dir = os.getcwd()

    while True:
        file_path = os.path.join(current_dir, filename)
        if os.path.exists(file_path):
            return os.path.abspath(file_path)

        # Move to the parent directory
        parent_dir = os.path.dirname(current_dir)

        # If we've reached the root directory and haven't found the file, stop
        if parent_dir == current_dir:
            break

        current_dir = parent_dir

    # If we still haven't found the file, use the default
    default_path = os.path.join(os.environ['HOME'], filename)
    return default_path if os.path.exists(default_path) else None


# pylint: disable=too-many-instance-attributes
class GoFigr:
    """\
    The GoFigr client. Handles all communication with the API: authentication, figure creation and manipulation,
    sharing, retrieval of user information, etc.

    """

    @from_config_or_env("GF_", find_config())
    def __init__(self,
                 username=None,
                 password=None,
                 api_key=None,
                 url=API_URL,
                 authenticate=True,
                 workspace_id=None,
                 anonymous=False,
                 asset_log=None):
        """\

        :param username: username to connect with
        :param password: password for authentication
        :param api_key: API key for authentication (specify instead of username & password)
        :param url: API URL
        :param authenticate: whether to authenticate right away. If False, authentication will happen during
        the first request.
        :param workspace_id: workspace ID to use for data syncing. Defaults to primary workspace.
        :param anonymous: True for anonymous access. Default False.
        :param asset_log: log of assets referenced by this instance

        """
        self.service_url = url
        self.username = username
        self.password = password
        self.api_key = api_key
        self.anonymous = anonymous
        self.workspace_id = workspace_id
        self.asset_log = asset_log if asset_log is not None else {}
        self._primary_workspace = None

        # Tokens for JWT authentication
        self._access_token = None
        self._refresh_token = None

        if authenticate:
            self.authenticate()

        self._bind_models()
        self._bind_readers()

        self._sync = None

    @property
    def sync(self):
        """Returns the default AssetSync object"""
        if not self._sync:
            self._sync = AssetSync(self, workspace_id=self.workspace_id, asset_log=self.asset_log)

        return self._sync

    def open(self, *args, **kwargs):
        """Opens a file using the default DataSync object"""
        return self.sync.open(*args, **kwargs)

    @property
    def app_url(self):
        """Returns the URL to the GoFigr app"""
        return self.service_url.replace("api", "app").replace(":8000", ":5173")

    def _bind_models(self):
        """\
        Create instance-bound model classes, e.g. Workspace, Figure, etc. Each will internally
        store a reference to this GoFigr client -- that way we don't have to pass it around.

        :return: None
        """
        # pylint: disable=too-few-public-methods,protected-access
        for name, obj in globals().items():
            if inspect.isclass(obj) and issubclass(obj, ModelMixin):
                class _Bound(obj):
                    _gf = self

                clean_name = name.replace("gf_", "")
                _Bound.__name__ = f"GoFigr.{clean_name}"
                _Bound.__qualname__ = f"GoFigr.{clean_name}"
                _Bound._gofigr_type_name = clean_name

                setattr(self, name.replace("gf_", ""), _Bound)
            elif inspect.isclass(obj) and issubclass(obj, NestedMixin):
                # Nested mixins don't reference the GoFigr object, but they're exposed in the same way
                # for consistency.
                setattr(self, name, obj)

    @property
    def api_url(self):
        """\
        Full URL to the API endpoint.
        """
        return f"{self.service_url}/api/{API_VERSION}/"

    @property
    def jwt_url(self):
        """\
        Full URL to the JWT endpoint (for authentication).
        """
        return f"{self.service_url}/api/token/"

    @staticmethod
    def _is_expired_token(response):
        """\
        Checks whether a response failed due to an expired auth token.

        :param response: Response object
        :return: True if failed due to an expired token, False otherwise.
        """
        if response.status_code != HTTPStatus.UNAUTHORIZED:
            return False
        try:
            obj = response.json()
            return obj.get('code') == 'token_not_valid'
        except ValueError:
            return False

    def create_api_key(self, name, expiry=None, workspace=None):
        """\
        Creates an API key

        :param name: name of the key to create
        :param expiry: expiration date. If None, the key will not expire.
        :param workspace: workspace for which the key is to be valid. If None, key will have access to the same
               workspaces as the user.
        :return: ApiKey instance

        """
        if expiry is not None and expiry.tzinfo is None:
            expiry = expiry.astimezone()

        # pylint: disable=no-member
        return self.ApiKey(name=name, expiry=expiry, workspace=workspace).create()

    def list_api_keys(self):
        """Lists all API keys"""
        # pylint: disable=no-member
        return self.ApiKey().list()

    def get_api_key(self, api_id):
        """Gets information about a specific API key"""
        # pylint: disable=no-member
        return self.ApiKey(api_id=api_id).fetch()

    def revoke_api_key(self, api_id):
        """Revokes an API key"""
        # pylint: disable=no-member
        if isinstance(api_id, str):
            return self.ApiKey(api_id=api_id).delete(delete=True)
        else:
            return api_id.delete(delete=True)

    def _request(self, method, endpoint, throw_exception=True, expected_status=(HTTPStatus.OK, ),
                 absolute_url=False, **kwargs):
        """\
        Convenience function for making HTTP requests.

        :param method: one of Session methods: Session.get, Session.post, etc.
        :param endpoint: relative API endpoint
        :param throw_exception: whether to check response status against expected_status and throw an exception
        :param expected_status: list of acceptable response status codes
        :param absolute_url: if False (default), interpret the endpoint relative to the API URL. Otherwise assume
                             it's fully qualified.
        :param kwargs: extra params passed verbatim to method(...)
        :return: Response

        """
        # pylint: disable=too-many-branches
        if not absolute_url:
            url = urljoin(self.api_url, endpoint)
        else:
            url = endpoint

        if not hasattr(expected_status, '__iter__'):
            expected_status = [expected_status, ]

        if self._access_token is None and self.api_key is None and not self.anonymous:
            raise RuntimeError("Please authenticate first")

        rqst = requests.session()
        try:
            if self.anonymous:
                response = method(rqst, url, **kwargs)
            elif self.api_key is None:
                response = method(rqst, url, headers={'Authorization': f'Bearer {self._access_token}'}, **kwargs)
            else:
                response = method(rqst, url, headers={'Authorization': f'Token {self.api_key}'}, **kwargs)

            if self._is_expired_token(response):
                self._refresh_access_token()
                return self._request(method, endpoint,
                                     throw_exception=throw_exception,
                                     expected_status=expected_status, **kwargs)

            if throw_exception and response.status_code not in expected_status:
                if response.status_code == HTTPStatus.FORBIDDEN:
                    raise UnauthorizedError(f"Unauthorized: {response.content}")
                elif response.status_code == HTTPStatus.METHOD_NOT_ALLOWED:
                    raise MethodNotAllowedError(f"Method not allowed: {response.content}")
                else:
                    raise RuntimeError(f"Request to {url} returned {response.status_code}: {response.content}")

            return response
        finally:
            rqst.close()

    def _get(self, endpoint, throw_exception=True, **kwargs):
        return self._request(Session.get, endpoint, throw_exception=throw_exception, **kwargs)

    def _post(self, endpoint, json, throw_exception=True, **kwargs):
        return self._request(Session.post, endpoint, json=json, throw_exception=throw_exception, **kwargs)

    def _patch(self, endpoint, json, throw_exception=True, **kwargs):
        return self._request(Session.patch, endpoint, json=json, throw_exception=throw_exception, **kwargs)

    def _put(self, endpoint, json, throw_exception=True, **kwargs):
        return self._request(Session.put, endpoint, json=json, throw_exception=throw_exception, **kwargs)

    def _delete(self, endpoint, throw_exception=True, **kwargs):
        return self._request(Session.delete, endpoint, throw_exception=throw_exception,
                             expected_status=HTTPStatus.NO_CONTENT, **kwargs)

    def heartbeat(self, throw_exception=True):
        """\
        Checks whether we can communicate with the API. Currently, this works by polling /api/v1/info.

        :param throw_exception: throw an exception if response code is not 200
        :return: Response

        """
        return self._get("info/", throw_exception=throw_exception)

    def _refresh_access_token(self):
        """\
        Refresh the JWT access token. If a refresh is not possible (e.g. the token has expired), will attempt
        to re-authenticate.

        :return: True if successful. Exception if not.
        """
        rqst = requests.session()
        try:
            rsp = rqst.post(self.jwt_url + "refresh/",
                            data={'refresh': self._refresh_token},
                            allow_redirects=False)

            if rsp.status_code == 200:
                self._access_token = rsp.json()['access']
                return True
            else:
                return self.authenticate()
        finally:
            if rqst is not None:
                rqst.close()

    def _authenticate_jwt(self):
        rqst = requests.session()
        try:
            rsp = rqst.post(self.jwt_url,
                            data={'username': self.username, 'password': self.password},
                            allow_redirects=False)

            if rsp.status_code != 200:
                raise RuntimeError("Authentication failed")

            self._refresh_token = rsp.json()['refresh']
            self._access_token = rsp.json()['access']
            return True
        finally:
            if rqst is not None:
                rqst.close()

    def authenticate(self):
        """\
        Authenticates with the API.

        :return: True
        """
        if self.anonymous:
            self.username = None
            return True
        elif self.api_key is not None:
            # With an API key there's no separate auth step, so we make sure everything works by querying user info
            info = self.user_info()
            self.username = info.username
            return True
        else:
            return self._authenticate_jwt()

    def _find_workspace_by_name(self, name, create, description=None):
        """\
        Finds a workspace by name.

        :param name: name of the workspace
        :param create: whether to create a new workspace or raise an exception if one doesn't exist
        :param description: optional description of the workspace
        :return: Workspace object

        """
        matches = [wx for wx in self.workspaces if wx.name == name]
        if len(matches) == 0:
            if create:
                wx = self.Workspace(name=name, description=description)
                wx.create()
                print(f"Created a new workspace: {wx.api_id}")
                return wx
            else:
                raise RuntimeError(f'Could not find workspace named "{name}"')
        elif len(matches) > 1:
            raise RuntimeError(f'Multiple (n={len(matches)}) workspaces match name "{name}". '
                               f'Please use an API ID instead.')
        else:
            return matches[0]

    def find_analysis(self, workspace, query):
        """\
        Finds an analysis within a workspace

        :param workspace: parent workspace (a gf.Workspace object)
        :param query: gf.Analysis, UUID string, ApiId, or FindByName
        :return: gf.Analysis object

        """
        api_id = try_parse_uuid4(query)

        if query is None:
            raise ValueError("Please specify an analysis")
        elif workspace is None:
            raise ValueError("Please specify a workspace")
        elif isinstance(query, gf_Analysis):
            return query
        elif isinstance(query, NotebookName):
            return query  # will be set by the Jupyter extension
        elif isinstance(query, str):
            if api_id is not None:
                return self.Analysis(api_id=api_id)
            else:
                return workspace.get_analysis(name=query, description="", create=True)
        elif isinstance(query, ApiId):
            return self.Analysis(api_id=query.api_id)
        elif isinstance(query, FindByName):
            if workspace.analyses is None:
                workspace.fetch()

            return workspace.get_analysis(name=query.name, description=query.description, create=query.create)
        else:
            raise ValueError(f"Unsupported query type {query}")

    def find_figure(self, analysis, query):
        """\
        Finds a figure within an analysis

        :param analysis: parent analysis (a gf.Analysis object)
        :param query: gf.Figure, UUID string, ApiId, or FindByName
        :return: gf.Figure object

        """
        api_id = try_parse_uuid4(query)

        if query is None:
            raise ValueError("Please specify a query")
        elif analysis is None:
            raise ValueError("Please specify an analysis")
        elif isinstance(query, gf_Figure):
            return query
        elif isinstance(query, str):
            if api_id is not None:
                return self.Figure(api_id=api_id)
            else:
                return analysis.get_figure(name=query, description="", create=True)
        elif isinstance(query, ApiId):
            return self.Figure(api_id=query.api_id)
        elif isinstance(query, FindByName):
            if analysis.figures is None:
                analysis.fetch()

            return analysis.get_figure(name=query.name, description=query.description, create=query.create)
        else:
            raise ValueError(f"Unsupported query type {query}")

    def find_workspace(self, query): # pylint: disable=too-many-return-statements
        """\
        Finds a workspace.

        :param query: gf.Workspace, UUID string, ApiId, or FindByName
        :return: gf.Workspace object

        """
        api_id = try_parse_uuid4(query)

        if query is None:
            # Use default workspace
            if self.primary_workspace is not None:
                return self.primary_workspace
            elif len(self.workspaces) == 1:  # this will happen if we're using a scoped API token
                return self.workspaces[0]
            else:
                raise ValueError("Please specify a workspace")
        elif isinstance(query, gf_Workspace):
            return query
        elif isinstance(query, str):
            if api_id is not None:
                return self.Workspace(api_id=query)
            else:
                return self._find_workspace_by_name(query, create=True)
        elif isinstance(query, ApiId):
            return self.Workspace(api_id=query.api_id)
        elif isinstance(query, FindByName):
            return self._find_workspace_by_name(query.name, query.create, description=query.description)
        else:
            raise ValueError(f"Unsupported query type {query}")


    def user_info(self, username=None):
        """\
        Retrieves information about a user.

        :param username: username. Set to None for self.
        :return: UserInfo object.

        """
        if not username:
            return UserInfo.from_json(self._get("user").json()[0])
        else:
            return UserInfo.from_json(self._get("user/" + username).json())

    def update_user_info(self, user_info, username=None):
        """\
        Updates user information for a user.

        :param user_info: UserInfo instance
        :param username: optional username. This is for testing only -- you will get an error if attempting \
        to update information for anybody other than yourself.
        :return: refreshed UserInfo from server

        """
        response = self._put("user/" + (username or user_info.username) + "/", user_info.to_json())
        return UserInfo.from_json(response.json())

    @property
    def workspaces(self):
        """Returns a list of all workspaces that the current user is a member of."""
        # pylint: disable=no-member
        return self.Workspace.list()

    @property
    def organizations(self):
        """Returns a list of all organizations that the current user is a member of."""
        # pylint: disable=no-member
        return self.Organization.list()

    @property
    def primary_workspace(self):
        """\
        Returns the primary workspace for this user.

        :return: Workspace instance

        """
        if self._primary_workspace is not None:
            return self._primary_workspace

        primaries = [w for w in self.workspaces if w.workspace_type == "primary"]
        primaries = [w for w in primaries if any(wm.username == self.username \
                                                 and wm.membership_type == WorkspaceMembership.OWNER
                                                 for wm in w.get_members(unauthorized_error=False))]

        if self.api_key is not None and len(primaries) == 0:
            self._primary_workspace = None
            return self._primary_workspace

        pw = assert_one(primaries,
                        "No primary workspace found. Please contact support.",
                        "Multiple primary workspaces found. Please contact support.")

        self._primary_workspace = pw
        return self._primary_workspace

    def _bind_readers(self):
        def _bind_one(name):
            # pylint: disable=unnecessary-lambda
            setattr(self, name, lambda *args, **kwargs: getattr(self.sync, name)(*args, **kwargs))

        for name in PANDAS_READERS:
            _bind_one(name)


def load_ipython_extension(ip):
    """\
    Loads the Jupyter extension. Present here so that we can do "%load_ext gofigr" without having to refer
    to a subpackage.

    :param ip: IPython shell
    :return: None

    """
    # pylint: disable=import-outside-toplevel
    from gofigr.jupyter import _load_ipython_extension
    return _load_ipython_extension(ip)


class AssetSync:
    """Provides drop-in replacements for open, read_xlsx, read_csv which version the data with the GoFigr service."""

    def __init__(self, gf, workspace_id=None, asset_log=None):
        """\

        :param gf: GoFigr instance
        :param workspace_id: workspace to sync under
        :param asset_log: dictionary of data revision IDs -> data revision objects
        """
        self.gf = gf
        self.workspace_id = workspace_id or self.gf.primary_workspace.api_id
        self.asset_log = asset_log if asset_log is not None else {}
        self._bind_readers()

        logging.debug(f"Using workspace ID {self.workspace_id}")

    @property
    def revisions(self):
        """\
        Returns all revisions in the log.
        """
        return self.asset_log.values()

    def clear_revisions(self):
        """\
        Clears the revision log
        """
        self.asset_log.clear()

    def _new_asset(self, pathlike):
        """\
        Creates a new asset from the given pathlike object.

        :param pathlike: local path to the asset e.g. ~/test.txt
        :return: Asset instance

        """
        logging.debug(f"Creating new asset for {pathlike}")
        ds = self.gf.Asset(name=os.path.basename(pathlike), workspace=self.gf.Workspace(api_id=self.workspace_id))
        ds.create()
        logging.debug(f"Created asset {ds.api_id}")
        return ds

    def _new_revision(self, pathlike):
        """\
        Creates a new revision from the given pathlike object. The revision will be created under an
        existing Asset if one with the same basename already eixsts. Otherwise, a new asset will be created.

        """
        logging.debug("New revision detected. Syncing...")
        assets = self.gf.Asset.find_by_name(os.path.basename(pathlike))
        logging.debug(f"Found assets: {assets}")

        # First, figure out which asset we're syncing to
        if len(assets) == 0:
            ds = self._new_asset(pathlike)
        elif len(assets) == 1:
            ds = assets[0]
        else:
            warnings.warn(f"Multiple assets with the same name found. Defaulting to first: "
                          f"{[d.api_id for d in assets]}")
            ds = assets[0]

        logging.debug(f"Creating a new revision for asset {ds.api_id} with path {pathlike}")

        # Now create the revision under the asset
        rev = self.gf.AssetRevision(asset=ds,
                                    data=[self.gf.FileData.read(pathlike)]).create()
        return rev

    def _log(self, revision, is_new_revision=False):
        """\
        Stores a revision in the log.

        """
        revision.is_new_revision = is_new_revision
        self.asset_log[revision.asset.api_id] = revision  # only keep the latest revision for each asset
        logging.debug(f"Logged revision {revision.api_id} for asset {revision.asset.api_id}")
        logging.debug(f"Current revision cache: {self.asset_log.keys()}")
        return revision

    def sync_revision(self, pathlike):
        """\
        Syncs an asset: calculates the checksum for the file and either uploads it to GoFigr (if checksum isn't found)
        or returns the existing revision.

        :param pathlike: path to the file
        :return: AssetRevision instance

        """
        # Grab the checksum
        logging.debug(f"Syncing {pathlike}")

        checksum = self._calc_checksum(pathlike)
        if checksum is None:
            warnings.warn(f"Unable to calculate checksum for {pathlike}. Skipping sync.")
            return None

        logging.debug(f"Calculated checksum for {pathlike}: {checksum}")

        # Check if we already have this asset
        revisions = self.gf.AssetRevision.find_by_hash(checksum, "blake3")

        if len(revisions) == 0:
            return self._log(self._new_revision(pathlike), is_new_revision=True)
        elif len(revisions) == 1:
            logging.debug(f"Found existing revision {revisions[0].api_id}")
            return self._log(revisions[0])
        else:
            logging.debug(f"Found existing revisions: {[rev.api_id for rev in revisions]}")
            warnings.warn(f"Multiple assets with the same checksum found. Defaulting to first: "
                          f"{[d.api_id for d in revisions]}")
            return self._log(revisions[0])

    def __call__(self, *args, **kwargs):
        """Same as AssetSync.sync()"""
        return self.sync(*args, **kwargs)


    def sync(self, pathlike, quiet=False):
        """\
        Syncs an asset: calculates the checksum for the file and either uploads it to GoFigr (if checksum isn't found)
        or returns the existing revision.

       :param pathlike: path to the file
       :param quiet: suppress Jupyter widget output
       :return: pathlike

        """
        rev = self.sync_revision(pathlike)
        if rev:
            logging.info(f"Asset synced: {rev.app_url}")

            if not quiet:
                AssetWidget(rev).show()
        return pathlike


    @contextlib.contextmanager
    def open_and_get_revision(self, pathlike, *args, **kwargs):
        """Syncs the data at pathlike with GoFigr and returns a tuple of file handle, AssetRevision instance."""
        f = None
        try:
            rev = self.sync_revision(pathlike)
            if rev:
                logging.info(f"Asset synced: {rev.app_url}")
                AssetWidget(rev).show()

            f = open(pathlike, *args, **kwargs)  # pylint: disable=unspecified-encoding
            yield f, rev
        finally:
            if f is not None and not f.closed:
                f.close()

    @contextlib.contextmanager
    def open(self, pathlike, *args, **kwargs):
        """Syncs the data at pathlike with GoFigr and returns an open file handle. Drop-in replacement for open()."""
        with self.open_and_get_revision(pathlike, *args, **kwargs) as (f, _):
            yield f

    def _wrap_reader(self, func):
        """Wraps a pandas reader function (e.g. read_csv) to provide data versioning and sync."""
        def wrapper(pathlike, *args, **kwargs):
            logging.debug(f"Calling {func.__name__} for {pathlike}")
            with self.open_and_get_revision(pathlike, 'rb') as (f, rev):
                frame = func(f, *args, **kwargs)
                frame.attrs = {REVISION_ATTR: rev.api_id}
                return frame
        return wrapper

    def _calc_checksum(self, pathlike):
        """Calculates a checksum for a file"""
        try:
            path = os.fspath(pathlike)
            if os.path.exists(path):
                file_hasher = blake3(max_threads=blake3.AUTO)  # pylint: disable=not-callable
                file_hasher.update_mmap(path)
                return file_hasher.hexdigest()
            else:
                warnings.warn(
                    "Non-local paths aren't supported yet. Please consider submitting an issue here: "
                    "https://github.com/GoFigr/gofigr-python/issues")
                return None
        except TypeError:
            warnings.warn(
                "This type of input isn't supported yet. Please consider submitting an issue here: "
                "https://github.com/GoFigr/gofigr-python/issues")
            return None

    def _bind_readers(self):
        """Binds all supported pandas reader functions."""
        for name in PANDAS_READERS:
            setattr(self, name, self._wrap_reader(getattr(pd, name)))



class NotebookName:
    """\
    Used as argument to configure() to specify that we want the analysis name to default to the name of the notebook
    """
    def __repr__(self):
        return "NotebookName"

ApiId = namedtuple("ApiId", ["api_id"])

class FindByName:
    """\
    Used as argument to configure() to specify that we want to find an analysis/workspace by name instead
    of using an API ID
    """
    def __init__(self, name, description=None, create=False):
        self.name = name
        self.description = description
        self.create = create

    def __repr__(self):
        return f"FindByName(name={self.name}, description={self.description}, create={self.create})"



"""\
Copyright (c) 2022-2025, Flagstaff Solutions, LLC
All rights reserved.

"""
from IPython import get_ipython


class GfDisplayPublisher:
    """\
    Custom IPython DisplayPublisher which traps all calls to publish() (e.g. when display(...) is called).

    """
    def __init__(self, pub, display_trap=None):
        """

        :param pub: Publisher to wrap around. We delegate all calls to this publisher unless trapped.
        :param display_trap: function to use as a display trap
        """
        self.pub = pub
        self.display_trap = display_trap

    def publish(self, data, *args, **kwargs):
        """
        IPython calls this method whenever it needs data displayed. Our function traps the call
        and calls DISPLAY_TRAP instead, giving it an option to suppress the figure from being displayed.

        We use this trap to publish the figure if auto_publish is True. Suppression is useful
        when we want to show a watermarked version of the figure, and prevents it from being showed twice (once
        with the watermark inside the trap, and once without in the originating call).

        :param data: dictionary of mimetypes -> data
        :param args: implementation-dependent
        :param kwargs: implementation-dependent
        :return: None

        """

        # Python doesn't support assignment to variables in closure scope, so we use a mutable list instead
        is_display_suppressed = [False]
        def suppress_display():
            is_display_suppressed[0] = True

        if self.display_trap is not None:
            trap = self.display_trap
            with SuppressDisplayTrap(self):
                trap(data, suppress_display=suppress_display)

        if not is_display_suppressed[0]:
            self.pub.publish(data, *args, **kwargs)

    def __getattr__(self, item):
        """\
        Delegates to self.pub

        :param item:
        :return:
        """
        if item == "pub":
            return super().__getattribute__(item)

        return getattr(self.pub, item)

    def __setattr__(self, key, value):
        """\
        Delegates to self.pub

        :param key:
        :param value:
        :return:
        """
        if key == "pub":
            super().__setattr__(key, value)

        return setattr(self.pub, key, value)

    def clear_output(self, *args, **kwargs):
        """IPython's clear_output. Defers to self.pub"""
        return self.pub.clear_output(*args, **kwargs)


class SuppressDisplayTrap:
    """\
    Context manager which temporarily suspends all display traps.
    """
    def __init__(self, gf_publisher=None):
        self.trap = None
        self.gf_publisher = gf_publisher or get_ipython().display_pub

    def __enter__(self):
        self.trap = self.gf_publisher.display_trap
        self.gf_publisher.display_trap = None

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.gf_publisher.display_trap = self.trap
        self.trap = None



"""\
Copyright (c) 2024, Flagstaff Solutions, LLC
All rights reserved.

"""
import inspect
import json
import logging
import os
import uuid
from base64 import b64encode
from importlib import resources

import six

from gofigr.databricks import get_config as get_databricks_config


def read_resource_text(package, resource):
    """\
    Reads a resource and returns it as a base-64 encoded string.

    :param package: package name
    :param resource: resource name
    :return: resource contents as a string

    """
    # pylint: disable=deprecated-method
    with resources.open_text(package, resource) as f:
        return f.read()


def read_resource_binary(package, resource):
    """\
    Reads a resource and returns it as bytes

    :param package: package name
    :param resource: resource name
    :return: bytes

    """
    # pylint: disable=deprecated-method
    with resources.open_binary(package, resource) as f:
        return f.read()


def read_resource_b64(package, resource):
    """\
    Reads a resource and returns it as a base-64 encoded string.

    :param package: package name
    :param resource: resource name
    :return: base64-encoded string

    """
    # pylint: disable=deprecated-method
    with resources.open_binary(package, resource) as f:
        return b64encode(f.read()).decode('ascii')


def from_config_or_env(env_prefix, config_path):
    """\
    Decorator that binds function arguments in order of priority (most important first):
    1. args/kwargs
    2. environment variables
    3. vendor-specific secret manager
    4. config file
    5. function defaults

    :param env_prefix: prefix for environment variables. Variables are assumed to be named \
    `<prefix> + <name of function argument in all caps>`, e.g. if prefix is ``MYAPP`` and function argument \
    is called host_name, we'll look for an \
    environment variable named ``MYAPP_HOST_NAME``.
    :param config_path: path to the JSON config file. If a list, will be checked in order until one is found.
    Function arguments will be looked up using their verbatim names.
    :return: decorated function

    """
    def decorator(func):
        @six.wraps(func)
        def wrapper(*args, **kwargs):
            # Read config file, if it exists
            if config_path and os.path.exists(config_path):
                logging.debug("Loading config from %s", config_path)
                with open(config_path, 'r', encoding='utf-8') as f:
                    try:
                        config_file = json.load(f)
                    except Exception as e:
                        raise RuntimeError(f"Error parsing configuration file {config_path}") from e
            else:
                config_file = {}

            dbconfig = get_databricks_config() or {}

            sig = inspect.signature(func)
            param_values = sig.bind_partial(*args, **kwargs).arguments
            for param_name in sig.parameters:
                env_name = f'{env_prefix}{param_name.upper()}'
                if param_name in param_values:
                    continue  # value supplied through args/kwargs: ignore env variables and the config file.
                elif env_name in os.environ:
                    param_values[param_name] = os.environ[env_name]
                elif param_name in dbconfig:
                    param_values[param_name] = dbconfig[param_name]
                elif param_name in config_file:
                    param_values[param_name] = config_file[param_name]

            return func(**param_values)

        return wrapper

    return decorator


def try_parse_uuid4(text):
    """Tries to parse a UUID from a string, returning it if successful, or returning None if not."""
    if text is None:
        return None

    try:
        return str(uuid.UUID(text, version=4))
    except Exception:  # pylint: disable=broad-exception-caught
        return None



import seaborn as sns
import matplotlib.pyplot as plt
from gofigr.publisher import Publisher

print(__file__)

# --- Setup GoFigr Publisher ---
# Initialize the publisher, specifying your workspace and analysis name.
# clear=True will remove previous figures in the analysis on each run.
pub = Publisher(workspace="Testz", analysis="Penguin Analysis", clear=True)

# --- Load Data ---
# Load the built-in penguins dataset
penguins = sns.load_dataset("penguins")

# --- 1. Scatter Plot ---
print("Generating and publishing scatter plot...")
sns.scatterplot(data=penguins, x="flipper_length_mm", y="bill_length_mm", hue="species")
plt.title("Penguin Bill Length vs. Flipper Length")
pub.publish(plt.gcf()) # Publish the current figure
plt.show()
plt.close() # Close the figure to start fresh for the next one
